import numpy as np

import astropy.units as u
from astropy.wcs.utils import pixel_to_pixel
from astropy.visualization import ImageNormalize, LinearStretch, PercentileInterval
from glue.core.link_helpers import LinkSame
from glue_jupyter.bqplot.image import BqplotImageView

from jdaviz.configs.imviz import wcs_utils
from jdaviz.core.astrowidgets_api import AstrowidgetsImageViewerMixin
from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import viewer_registry
from jdaviz.core.freezable_state import FreezableBqplotImageViewerState
from jdaviz.configs.default.plugins.viewers import JdavizViewerMixin
from jdaviz.utils import (get_wcs_only_layer_labels, data_has_valid_wcs,
                          layer_is_image_data, get_top_layer_index)

__all__ = ['ImvizImageView']


@viewer_registry("imviz-image-viewer", label="Image 2D (Imviz)")
class ImvizImageView(JdavizViewerMixin, BqplotImageView, AstrowidgetsImageViewerMixin):
    # categories: zoom resets, zoom, pan, subset, select tools, shortcuts
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:boxzoommatch', 'jdaviz:boxzoom'],
                    ['jdaviz:panzoommatch', 'jdaviz:imagepanzoom'],
                    ['bqplot:truecircle', 'bqplot:rectangle', 'bqplot:ellipse',
                     'bqplot:circannulus'],
                    ['jdaviz:blinkonce', 'jdaviz:contrastbias', 'jdaviz:selectcatalog'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export', 'jdaviz:sidebar_compass']
                ]

    default_class = None
    _state_cls = FreezableBqplotImageViewerState

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # provide reference from state back to viewer to use for zoom syncing
        self.state._set_viewer(self)
        self.init_astrowidgets_api()
        self._subscribe_to_layers_update()

        self.compass = None
        self.line_profile_xy = None

        self.add_event_callback(self.on_mouse_or_key_event, events=['keydown'])
        self.state.add_callback('x_min', self.on_limits_change)
        self.state.add_callback('x_max', self.on_limits_change)
        self.state.add_callback('y_min', self.on_limits_change)
        self.state.add_callback('y_max', self.on_limits_change)

        self.state.show_axes = False
        self.figure.fig_margin = {'left': 0, 'bottom': 0, 'top': 0, 'right': 0}

        # By default, glue computes a fixed resolution buffer that matches the
        # axes - but this means that when panning, one sees white outside of
        # the original buffer until the buffer updates again, thus there is a
        # lag in the image display. By increasing the external padding to 0.5
        # the image is made larger by 50% along all four sides, helping create
        # the illusion of smooth panning. We can increase this further to
        # improve the panning experience, but this can cause a larger delay
        # when the image does need to update as it will be more computationally
        # intensive.
        self.state.image_external_padding = 0.5

        self.data_menu._obj.dataset.add_filter('is_image')

    def on_mouse_or_key_event(self, data):
        active_image_layer = self.active_image_layer
        if active_image_layer is None:
            return

        if self.line_profile_xy is None:
            try:
                self.line_profile_xy = self.session.jdaviz_app.get_tray_item_from_name(
                    'imviz-line-profile-xy')
            except KeyError:  # pragma: no cover
                return

        if data['event'] == 'keydown':
            key_pressed = data['key']

            if key_pressed in ('b', 'B'):
                self.blink_once(reversed=key_pressed == 'B')

    def blink_once(self, reversed=False):
        # Simple blinking of images - this will make it so that only one
        # layer is visible at a time and cycles through the layers.

        # Exclude Subsets (they are global) and children via associated data

        def is_parent(data):
            return self.session.jdaviz_app._get_assoc_data_parent(data.label) is None

        valid = [ilayer for ilayer, layer in enumerate(self.state.layers)
                 if layer_is_image_data(layer.layer) and is_parent(layer.layer)]
        children = [ilayer for ilayer, layer in enumerate(self.state.layers)
                    if layer_is_image_data(layer.layer) and not is_parent(layer.layer)]

        n_layers = len(valid)

        if n_layers == 1:
            msg = SnackbarMessage(
                'Nothing to blink. Select a second image in the Data menu to use this feature.',
                color='warning', sender=self)
            self.session.hub.broadcast(msg)

        elif n_layers > 1:
            # If only one layer is visible, pick the next one to be visible,
            # otherwise start from the last visible one.

            visible = [ilayer for ilayer in valid if self.state.layers[ilayer].visible]
            n_visible = len(visible)

            if n_visible == 0:
                msg = SnackbarMessage('No visible layer to blink',
                                      color='warning', sender=self)
                self.session.hub.broadcast(msg)
            elif n_visible > 0:
                if not reversed:
                    delta = 1
                else:
                    delta = -1
                next_layer = valid[(valid.index(visible[-1]) + delta) % n_layers]
                self.state.layers[next_layer].visible = True

                # make invisible all parent layers other than the next layer:
                layers_to_set_not_visible = set(valid) - set([next_layer])
                # no child layers are visible by default:
                layers_to_set_not_visible.update(set(children))

                for ilayer in layers_to_set_not_visible:
                    self.state.layers[ilayer].visible = False

                # We can display the active data label in Compass plugin.
                self.set_compass(self.state.layers[next_layer].layer)

                # Update line profile plots too.
                if self.line_profile_xy is None:
                    try:
                        self.line_profile_xy = self.session.jdaviz_app.get_tray_item_from_name(
                            'imviz-line-profile-xy')
                    except KeyError:  # pragma: no cover
                        return
                self.line_profile_xy.viewer_selected = self.reference_id
                self.line_profile_xy.vue_draw_plot()

    def on_limits_change(self, *args):
        try:
            i = get_top_layer_index(self)
        except IndexError:
            if self.compass is not None:
                self.compass.clear_compass()
            return
        if i is None:
            return
        self.set_compass(self.state.layers[i].layer)

    @property
    def top_visible_data_label(self):
        """Data label of the top visible layer in the viewer."""
        try:
            i = get_top_layer_index(self)
        except IndexError:
            data_label = ''
        else:
            if i is None:
                data_label = ''
            else:
                data_label = self.state.layers[i].layer.label
        return data_label

    @property
    def first_loaded_data(self):
        """Data that is first loaded into the viewer.
        This may not be the visible layer.
        Returns `None` if no real data is loaded.
        """
        for lyr in self.layers:
            data = lyr.layer
            if layer_is_image_data(data):
                return data

    def _get_real_xy(self, image, x, y, reverse=False):
        """Return real (X, Y) position and status in case of dithering as well as whether the
        results were within the bounding box of the reference data or required possibly inaccurate
        extrapolation.

        ``coords_status`` is for ``CoordsInfo`` coords handling only.
        When `True`, it sets the coords, otherwise it resets.

        ``reverse=True`` is only for internal roundtripping (e.g., centroiding
        in Subset Tools plugin). Never use this for coordinates display panel.

        """
        # By default we'll assume the coordinates are valid and within any applicable bounding box.
        unreliable_world = False
        unreliable_pixel = False
        if data_has_valid_wcs(image):
            # Convert these to a SkyCoord via WCS - note that for other datasets
            # we aren't actually guaranteed to get a SkyCoord out, just for images
            # with valid celestial WCS
            try:
                align_by = self.get_alignment_method(image.label).lower()

                # Convert X,Y from reference data to the one we are actually seeing.
                # world_to_pixel return scalar ndarray that we need to convert to float.
                if align_by == 'wcs':
                    if not reverse:
                        outside_ref_bounding_box = wcs_utils.data_outside_gwcs_bounding_box(
                            self.state.reference_data, x, y)
                        x, y = list(map(float, pixel_to_pixel(
                            self.state.reference_data.coords, image.coords, x, y)))
                        outside_image_bounding_box = wcs_utils.data_outside_gwcs_bounding_box(
                            image, x, y)
                        unreliable_pixel = outside_image_bounding_box or outside_ref_bounding_box
                        unreliable_world = unreliable_pixel
                    else:
                        # We don't bother with unreliable_pixel and unreliable_world computation
                        # because this takes input (x, y) in the frame of visible layer and wants
                        # to convert it back to the frame of reference layer to pass back to the
                        # viewer. At this point, we no longer know if input (x, y) is accurate
                        # or not.
                        x, y = list(map(float, pixel_to_pixel(
                            image.coords, self.state.reference_data.coords, x, y)))
                else:  # pixels or self
                    unreliable_world = wcs_utils.data_outside_gwcs_bounding_box(image, x, y)

                coords_status = True
            except Exception:
                coords_status = False
        else:
            coords_status = False

        return x, y, coords_status, (unreliable_world, unreliable_pixel)

    def _get_zoom_limits(self, image):
        """Return a list of ``(x, y)`` that defines four corners of
        the zoom box for a given image.
        This is needed because viewer values are only based on reference
        image, which can be inaccurate if given image is dithered and
        they are linked by WCS.
        """
        if self.state.reference_data.meta.get('_WCS_ONLY', False):
            corner_world_coords = self.state.reference_data.coords.pixel_to_world(
                (self.state.x_min, self.state.x_min, self.state.x_max, self.state.x_max),
                (self.state.y_min, self.state.y_max, self.state.y_max, self.state.y_min)
            )
            # Convert X,Y from reference data to the one we are actually seeing.
            x = image.coords.world_to_pixel(corner_world_coords)
            zoom_limits = np.array(list(zip(x[0], x[1])))
        else:
            zoom_limits = np.array(((self.state.x_min, self.state.y_min),
                                    (self.state.x_min, self.state.y_max),
                                    (self.state.x_max, self.state.y_max),
                                    (self.state.x_max, self.state.y_min)))

        return zoom_limits

    def set_compass(self, image):
        """Update the Compass plugin with info from the given image Data object."""
        if self.compass is None:  # Maybe another viewer has it
            return

        zoom_limits = self._get_zoom_limits(image)

        # Downsample input data to about 400px (as per compass.vue) for performance.
        xstep = max(1, round(image.shape[1] / 400))
        ystep = max(1, round(image.shape[0] / 400))
        arr = image[image.main_components[0]][::ystep, ::xstep]
        vmin, vmax = PercentileInterval(95).get_limits(arr)
        norm = ImageNormalize(vmin=vmin, vmax=vmax, stretch=LinearStretch())
        self.compass.draw_compass(image.label, wcs_utils.draw_compass_mpl(
            arr, orig_shape=image.shape, wcs=image.coords, show=False, zoom_limits=zoom_limits,
            norm=norm))

    def set_plot_axes(self):
        self.figure.axes[1].tick_format = None
        self.figure.axes[0].tick_format = None

        self.figure.axes[1].label = "y: pixels"
        self.figure.axes[0].label = "x: pixels"

        # Make it so y axis label is not covering tick numbers.
        self.figure.axes[1].label_offset = "-50"

    def data(self, cls=None):
        return [layer_state.layer  # .get_object(cls=cls or self.default_class)
                for layer_state in self.state.layers
                if hasattr(layer_state, 'layer') and
                layer_is_image_data(layer_state.layer)]

    def get_alignment_method(self, data_label):
        """Find the type of ``glue`` linking between the given
        data label with the reference data in viewer.

        Parameters
        ----------
        data_label : str
            Data label to look up.

        Returns
        -------
        align_by : {'pixels', 'wcs', 'self'}
            One of the link types accepted by the Orientation plugin
            or ``'self'`` if the data label belongs to the reference data itself.

        Raises
        ------
        ValueError
            Link look-up failed.

        """
        if len(self.session.application.data_collection) == 0:
            raise ValueError('No reference data for link look-up')

        ref_label = getattr(self.state.reference_data, 'label', None)
        if data_label == ref_label:
            return 'self'

        if ref_label in get_wcs_only_layer_labels(self.jdaviz_app):
            return 'wcs'

        align_by = None
        for elink in self.session.application.data_collection.external_links:
            elink_labels = (elink.data1.label, elink.data2.label)
            if data_label in elink_labels and ref_label in elink_labels:
                if isinstance(elink, LinkSame):  # Assumes WCS link never uses LinkSame
                    align_by = 'pixels'
                else:  # If not pixels, must be WCS
                    align_by = 'wcs'
                break  # Might have duplicate, just grab first match

        if align_by is None:
            raise ValueError(f'{data_label} not found in data collection external links')

        return align_by

    def _get_fov(self, wcs=None):
        if wcs is None:
            wcs = self.state.reference_data.coords
        if self.jdaviz_app._align_by != "wcs" or wcs is None:
            return

        # compute the mean of the height and width of the
        # viewer's FOV on ``data`` in world units:
        x_corners = [
            self.state.x_min,
            self.state.x_max,
            self.state.x_min
        ]
        y_corners = [
            self.state.y_min,
            self.state.y_min,
            self.state.y_max
        ]

        sky_corners = wcs.pixel_to_world(x_corners, y_corners)
        height_sky = abs(sky_corners[0].separation(sky_corners[2]))
        width_sky = abs(sky_corners[0].separation(sky_corners[1]))
        fov_sky = u.Quantity([height_sky, width_sky]).mean()
        return fov_sky

    def _get_center_skycoord(self, data=None):
        # get SkyCoord for the center of ``data`` in this viewer:
        x_cen = (self.state.x_min + self.state.x_max) * 0.5
        y_cen = (self.state.y_min + self.state.y_max) * 0.5

        if (self.jdaviz_app._align_by == "wcs" or data is None
                or data.label == self.state.reference_data.label):
            return self.state.reference_data.coords.pixel_to_world(x_cen, y_cen)

        if data.coords is not None:
            return data.coords.pixel_to_world(x_cen, y_cen)



import math
import numpy as np
from traitlets import Bool, Unicode, observe

from astropy import units as u
from bqplot import LinearScale
from glue.core import BaseData
from glue_jupyter.bqplot.image.layer_artist import BqplotImageSubsetLayerArtist

from jdaviz.configs.cubeviz.plugins.viewers import CubevizImageView
from jdaviz.configs.imviz.plugins.viewers import ImvizImageView
from jdaviz.configs.mosviz.plugins.viewers import (MosvizImageView, MosvizProfileView,
                                                   MosvizProfile2DView)
from jdaviz.configs.rampviz.plugins.viewers import RampvizImageView, RampvizProfileView
from jdaviz.configs.specviz.plugins.viewers import SpecvizProfileView
from jdaviz.core.custom_units_and_equivs import PIX2
from jdaviz.core.events import ViewerAddedMessage, GlobalDisplayUnitChanged
from jdaviz.core.helpers import data_has_valid_wcs
from jdaviz.core.marks import PluginScatter, PluginLine
from jdaviz.core.registries import tool_registry
from jdaviz.core.template_mixin import TemplateMixin, DatasetSelectMixin
from jdaviz.core.unit_conversion_utils import (all_flux_unit_conversion_equivs,
                                               check_if_unit_is_per_solid_angle,
                                               flux_conversion_general)

__all__ = ['CoordsInfo']


@tool_registry('g-coords-info')
class CoordsInfo(TemplateMixin, DatasetSelectMixin):
    template_file = __file__, "coords_info.vue"

    _supported_viewer_classes = (SpecvizProfileView,
                                 ImvizImageView,
                                 CubevizImageView,
                                 RampvizImageView,
                                 RampvizProfileView,
                                 MosvizImageView,
                                 MosvizProfile2DView)

    _viewer_classes_with_marker = (RampvizProfileView, SpecvizProfileView, MosvizProfile2DView)

    dataset_icon = Unicode("").tag(
        sync=True
    )  # option for layer (auto, none, or specific layer)
    icon = Unicode("").tag(sync=True)  # currently exposed layer

    row1a_title = Unicode("").tag(sync=True)
    row1a_text = Unicode("").tag(sync=True)
    row1b_title = Unicode("").tag(sync=True)
    row1b_text = Unicode("").tag(sync=True)
    row1_unreliable = Bool(False).tag(sync=True)

    row2_title = Unicode("").tag(sync=True)
    row2_text = Unicode("").tag(sync=True)
    row2_unreliable = Bool(False).tag(sync=True)

    row3_title = Unicode("").tag(sync=True)
    row3_text = Unicode("").tag(sync=True)
    row3_unreliable = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._marks = {}
        self._dict = {}  # dictionary representation of current mouseover info
        self._x, self._y = None, None  # latest known cursor positions
        self.image_unit = None

        # subscribe/unsubscribe to mouse events across all existing viewers
        viewer_refs = []
        for viewer in self.app._viewer_store.values():
            if isinstance(viewer, self._supported_viewer_classes):
                self._create_viewer_callbacks(viewer)
                viewer_refs.append(viewer.reference_id)

        self.dataset._manual_options = ['auto', 'none']

        self.dataset.filters = ['layer_in_viewers', 'is_not_wcs_only', 'layer_is_not_dq']
        if self.app.config == 'imviz':
            # filter out scatter-plot entries (from add_markers API, for example)
            self.dataset.add_filter('is_image')

        # subscribe to mouse events on any new viewers
        self.hub.subscribe(self, ViewerAddedMessage, handler=self._on_viewer_added)
        if self.config == "cubeviz":
            self.hub.subscribe(
                self, GlobalDisplayUnitChanged, handler=self._on_global_display_unit_changed
            )

    def _create_marks_for_viewer(self, viewer, id=None):
        if id is None:
            id = viewer.reference_id
        if id in self._marks:
            return
        if isinstance(viewer, MosvizProfile2DView):
            self._marks[id] = PluginLine(viewer,
                                         x=[0, 0], y=[0, 1],
                                         scales={'x': viewer.scales['x'],
                                                 'y': LinearScale(min=0, max=1)},
                                         visible=False)
        else:
            self._marks[id] = PluginScatter(viewer,
                                            marker='rectangle', stroke_width=1,
                                            visible=False)
        if isinstance(viewer, MosvizProfileView):
            matched_id = f"{id}:matched"
            self._marks[matched_id] = PluginLine(viewer,
                                                 x=[0, 0], y=[0, 1],
                                                 scales={'x': viewer.scales['x'],
                                                         'y': LinearScale(min=0, max=1)},
                                                 visible=False)
            viewer.figure.marks = viewer.figure.marks + [self._marks[matched_id]]

        viewer.figure.marks = viewer.figure.marks + [self._marks[id]]

    def _create_viewer_callbacks(self, viewer):
        if isinstance(viewer, self._supported_viewer_classes):
            if isinstance(viewer, self._viewer_classes_with_marker):
                self._create_marks_for_viewer(viewer)
            callback = self._viewer_callback(viewer, self._viewer_mouse_event)
            viewer.add_event_callback(callback, events=['mousemove', 'mouseleave', 'mouseenter'])

            viewer.state.add_callback('layers', lambda msg: self._layers_changed(viewer))

    def _on_viewer_added(self, msg):
        self._create_viewer_callbacks(self.app.get_viewer_by_id(msg.viewer_id))

    def _on_global_display_unit_changed(self, msg):

        # all cubes are converted to surface brightness so we just need to
        # listen to SB for cubeviz unit changes
        if msg.axis == "sb":
            self.image_unit = u.Unit(msg.unit)

    @property
    def marks(self):
        """
        Access the marks created by this plugin.
        """
        if self._marks:
            # TODO: replace with cache property?
            return self._marks

        # create marks for each of the spectral viewers (will need a listener event to create marks
        # for new viewers if dynamic creation of spectral viewers is ever supported)
        for id, viewer in self.app._viewer_store.items():
            if isinstance(viewer, self._viewer_classes_with_marker):
                self._create_marks_for_viewer(viewer, id)
        return self._marks

    @property
    def _matched_markers(self):
        if self.app.config == 'specviz2d':
            return {'specviz2d-0': ['specviz2d-1:matched'],
                    'specviz2d-1': ['specviz2d-0']}
        if self.app.config == 'mosviz':
            return {'mosviz-1': ['mosviz-2:matched'],
                    'mosviz-2': ['mosviz-1']}
        return {}

    def as_text(self):
        return (f"{self.row1a_title} {self.row1a_text} {self.row1b_title} {self.row1b_text}".strip(),  # noqa
                f"{self.row2_title} {self.row2_text}".strip(),
                f"{self.row3_title} {self.row3_text}".strip())

    def as_dict(self):
        return self._dict

    def reset_coords_display(self):
        self.row1a_title = '\u00A0'  # to force empty line if no other content
        self.row1a_text = ""
        self.row1b_title = ""
        self.row1b_text = ""
        self.row1_unreliable = False

        self.row2_title = '\u00A0'
        self.row2_text = ""
        self.row2_unreliable = False

        self.row3_title = '\u00A0'
        self.row3_text = ""
        self.row3_unreliable = False

        self.icon = ""
        self._dict = {}

    def _viewer_mouse_clear_event(self, viewer, data=None):
        self.reset_coords_display()
        marker_ids = [viewer._reference_id] + self._matched_markers.get(viewer._reference_id, [])
        for marker_id in marker_ids:
            marks = self.marks.get(marker_id)
            if marks is not None:
                marks.visible = False
        self.app.state.show_toolbar_buttons = True

    def _viewer_mouse_event(self, viewer, data):
        if data['event'] in ('mouseleave', 'mouseenter'):
            self._viewer_mouse_clear_event(viewer, data)
            return

        if len(self.app.data_collection) < 1:
            self._viewer_mouse_clear_event(viewer)
            return

        # otherwise a mousemove event, we need to get cursor coordinates and update the display

        # Extract data coordinates - these are pixels in the reference image
        x = data['domain']['x']
        y = data['domain']['y']

        if x is None or y is None:  # Out of bounds
            self._viewer_mouse_clear_event(viewer)
            return

        # update last known cursor position (so another event like a change in layers can update
        # the coordinates with the last known position)
        self._x, self._y = x, y
        self.update_display(viewer, x=x, y=y)

    def _layers_changed(self, viewer):
        if self._x is None or self._y is None:
            return

        # update display for a (possible) change to the active layer based on the last known
        # cursor position
        self.update_display(viewer, self._x, self._y)

    @observe('dataset_selected')
    def _selected_dataset_changed(self, *args):
        if self.dataset_selected == 'auto':
            self.dataset_icon = 'mdi-auto-fix'
        elif self.dataset_selected == 'none':
            self.dataset_icon = 'mdi-cursor-default'
        else:
            self.dataset_icon = self.app.state.layer_icons.get(self.dataset_selected, '')

    def vue_next_layer(self, *args, **kwargs):
        self.dataset.select_next()

    def update_display(self, viewer, x, y):
        self._dict = {}
        if isinstance(viewer, (SpecvizProfileView, RampvizProfileView)):
            self._spectrum_viewer_update(viewer, x, y)
        elif isinstance(viewer,
                        (ImvizImageView, CubevizImageView,
                         MosvizImageView, MosvizProfile2DView,
                         RampvizImageView)
                        ):
            self._image_viewer_update(viewer, x, y)

    def _image_shape_inds(self, image):
        # return the indices in image.shape for the x and y dimension, respectively
        if image.ndim == 3:
            # cubeviz case
            return (0, 1)  # (ix_shape, iy_shape)
        elif image.ndim == 2:
            return (1, 0)  # (ix_shape, iy_shape)
        else:  # pragma: no cover
            raise ValueError(f'does not support ndim={image.ndim}')

    def _get_cube_value(self, image, arr, x, y, viewer):
        if image.ndim == 3:
            # cubeviz case:
            return arr[int(round(x)), int(round(y)), viewer.state.slices[-1]]
        elif image.ndim == 2:
            if isinstance(viewer, RampvizImageView):
                x, y = y, x
            return arr[int(round(y)), int(round(x))]
        else:  # pragma: no cover
            raise ValueError(f'does not support ndim={image.ndim}')

    def _image_viewer_update(self, viewer, x, y):
        # Display the current cursor coordinates (both pixel and world) as
        # well as data values. For now we use the first dataset in the
        # viewer for the data values.

        # Extract first dataset from visible layers and use this for coordinates - the choice
        # of dataset shouldn't matter if the datasets are linked correctly
        active_layer = viewer.active_image_layer
        if active_layer is None:
            self._viewer_mouse_clear_event(viewer)
            return

        self.app.state.show_toolbar_buttons = False
        if self.dataset.selected == 'auto':
            image = active_layer.layer
        elif self.dataset.selected == 'none':
            active_layer = viewer.layers[0].state
            image = viewer.layers[0].layer
        else:
            for layer in viewer.layers:
                if layer.layer.label == self.dataset.selected and layer.visible:
                    if isinstance(layer, BqplotImageSubsetLayerArtist):
                        # cannot expose info for spatial subset layers
                        continue
                    active_layer = layer.state
                    image = layer.layer
                    break
            else:
                image = None

        # If there is one, get the associated DQ layer for the active layer:
        associated_dq_layers = None
        available_plugins = [tray_item['name'] for tray_item in self.app.state.tray_items]
        if 'g-data-quality' in available_plugins:
            assoc_children = self.app._get_assoc_data_children(active_layer.layer.label)
            if assoc_children:
                data_quality_plugin = self.app.get_tray_item_from_name('g-data-quality')
                viewer_obj = self.app.get_viewer(viewer)
                associated_dq_layers = data_quality_plugin.get_dq_layers(viewer_obj)

        unreliable_pixel, unreliable_world = False, False

        self._dict['axes_x'] = x
        self._dict['axes_x:unit'] = 'pix'
        self._dict['axes_y'] = y
        self._dict['axes_y:unit'] = 'pix'

        # set default empty values
        if self.dataset.selected != 'none' and image is not None:
            self.icon = self.app.state.layer_icons.get(image.label, '')  # noqa
            self._dict['data_label'] = image.label

        # Separate logic for each viewer type, ultimately needs to result in extracting sky coords.
        # NOTE: pixel_to_world axes order is opposite of array value axes order, so...
        #   3D: pixel_to_world(z, y, x) -> arr[x, y, z]
        #   2D: pixel_to_world(x, y) -> arr[y, x]
        if self.dataset.selected == 'none' or image is None:
            self.icon = 'mdi-cursor-default'
            self._dict['data_label'] = ''
            coords_status = False

        elif isinstance(viewer, ImvizImageView):
            x, y, coords_status, (unreliable_world, unreliable_pixel) = viewer._get_real_xy(image, x, y)  # noqa
            if coords_status:
                try:
                    sky = image.coords.pixel_to_world(x, y).icrs
                except Exception:  # WCS might not be celestial
                    coords_status = False

        elif isinstance(viewer, CubevizImageView):
            # TODO: This assumes data_collection[0] is the main reference
            #  data for this application. This section will need to be updated
            #  when that is no longer true.
            # Hack to insert WCS for generated 2D and 3D images using FLUX cube WCS.
            if 'Plugin' in getattr(image, 'meta', {}) and not image.coords:
                coo_data = self.app.data_collection[0]
            else:
                coo_data = image

            if '_orig_spec' in getattr(coo_data, 'meta', {}):
                # Hack around various WCS propagation issues in Cubeviz, example:
                # https://github.com/glue-viz/glue-astronomy/issues/75
                data_wcs = coo_data.meta['_orig_spec'].wcs
                wcs_ndim = 3
            elif data_has_valid_wcs(coo_data):
                data_wcs = coo_data.coords
                wcs_ndim = coo_data.ndim
            else:
                data_wcs = None

            if data_wcs:
                try:
                    if wcs_ndim == 3:
                        sky = data_wcs.pixel_to_world(viewer.slice, y, x)[1].icrs
                    else:  # wcs_ndim == 2
                        sky = data_wcs.pixel_to_world(x, y).icrs
                except Exception:
                    coords_status = False
                else:
                    coords_status = True
            else:
                self.reset_coords_display()
                coords_status = False

            slice_plugin = self.app._jdaviz_helper.plugins.get('Slice', None)
            if slice_plugin is not None and len(image.shape) == 3:
                # float to be compatible with default value of nan
                self._dict['slice'] = float(viewer.slice)
                self._dict['spectral_axis'] = slice_plugin.value
                self._dict['spectral_axis:unit'] = slice_plugin._obj.value_unit

        elif isinstance(viewer, RampvizImageView):
            coords_status = False

            slice_plugin = self.app._jdaviz_helper.plugins.get('Slice', None)
            if slice_plugin is not None and len(image.shape) == 3:
                # float to be compatible with default value of nan
                self._dict['slice'] = float(viewer.slice)

        elif isinstance(viewer, MosvizImageView):

            if data_has_valid_wcs(image, ndim=2):
                try:
                    sky = image.coords.pixel_to_world(x, y).icrs
                except Exception:  # WCS might not be celestial  # pragma: no cover
                    coords_status = False
                else:
                    coords_status = True
            else:  # pragma: no cover
                self.reset_coords_display()
                coords_status = False

        elif isinstance(viewer, MosvizProfile2DView):
            self._dict['spectral_axis'] = self._dict['axes_x']
            self._dict['spectral_axis:unit'] = self._dict['axes_x:unit']
            self._dict['value'] = self._dict['axes_y']
            self._dict['value:unit'] = self._dict['axes_y:unit']
            coords_status = False

        if coords_status:
            celestial_coordinates = sky.to_string('hmsdms', precision=4, pad=True).split()
            celestial_coordinates_deg = sky.to_string('decimal', precision=10, pad=True).split()
            world_ra = celestial_coordinates[0]
            world_dec = celestial_coordinates[1]
            world_ra_deg = celestial_coordinates_deg[0]
            world_dec_deg = celestial_coordinates_deg[1]

            if "nan" in (world_ra, world_dec, world_ra_deg, world_dec_deg):
                self.reset_coords_display()

            self.row2_title = 'World'
            self.row2_text = f'{world_ra} {world_dec} (ICRS)'
            self.row2_unreliable = unreliable_world
            self.row3_title = ''
            self.row3_text = f'{world_ra_deg} {world_dec_deg} (deg)'
            self.row3_unreliable = unreliable_world
            self._dict['world_ra'] = sky.ra.value
            self._dict['world_dec'] = sky.dec.value
            self._dict['world:unreliable'] = unreliable_world
        elif isinstance(viewer, MosvizProfile2DView) and hasattr(getattr(image, 'coords', None),
                                                                 'pixel_to_world'):
            # use WCS to expose the wavelength for a 2d spectrum shown in pixel space
            try:
                wave, pixel = image.coords.pixel_to_world(x, y)
                if wave is not None:
                    equivalencies = all_flux_unit_conversion_equivs(cube_wave=wave)
                    wave = wave.to(self.app._get_display_unit('spectral'),
                                   equivalencies=equivalencies)
                    self._dict['spectral_axis'] = wave.value
                    self._dict['spectral_axis:unit'] = wave.unit.to_string()
            except Exception:  # WCS might not be valid  # pragma: no cover
                coords_status = False
            else:
                coords_status = True
                self.row2_title = 'Wave'
                self.row2_text = f'{wave.value:10.5e} {wave.unit.to_string()}'
                self.row2_unreliable = False

                self.row3_title = '\u00A0'
                self.row3_text = ""
                self.row3_unreliable = False
        else:
            self.row2_title = '\u00A0'
            self.row2_text = ""
            self.row2_unreliable = False

            self.row3_title = '\u00A0'
            self.row3_text = ""
            self.row3_unreliable = False

        maxsize = int(np.ceil(np.log10(np.max(active_layer.layer.shape)))) + 3
        fmt = 'x={0:0' + str(maxsize) + '.1f} y={1:0' + str(maxsize) + '.1f}'
        self.row1a_title = 'Pixel'
        self.row1a_text = (fmt.format(x, y))
        self.row1_unreliable = unreliable_pixel
        self._dict['pixel_x'] = float(x)
        self._dict['pixel_y'] = float(y)
        self._dict['pixel:unreliable'] = unreliable_pixel

        # Extract data values at this position.
        # TODO: for now we just use the first visible layer but we should think
        # of how to display values when multiple datasets are present.

        if self.dataset.selected == 'none' or image is None:
            # no data values to extract
            self.row1b_title = ''
            self.row1b_text = ''
            return

        # Extract data values at this position.
        # Check if shape is [x, y, z] or [y, x] and show value accordingly.
        ix_shape, iy_shape = self._image_shape_inds(image)

        if (-0.5 < x < image.shape[ix_shape] - 0.5 and -0.5 < y < image.shape[iy_shape] - 0.5
                and hasattr(active_layer, 'attribute')):

            attribute = active_layer.attribute

            if isinstance(viewer, (ImvizImageView, MosvizImageView, MosvizProfile2DView)):
                value = image.get_data(attribute)[int(round(y)), int(round(x))]

                if associated_dq_layers is not None:
                    associated_dq_layer = associated_dq_layers[0]
                    dq_attribute = associated_dq_layer.state.attribute
                    dq_data = associated_dq_layer.layer.get_data(dq_attribute)
                    dq_value = dq_data[int(round(y)), int(round(x))]

                unit = u.Unit(image.get_component(attribute).units)
                if (isinstance(viewer, MosvizProfile2DView) and unit != ''
                   and u.Unit(self.app._get_display_unit(attribute)).physical_type
                   not in ['frequency', 'wavelength', 'length']
                   and unit != self.app._get_display_unit(attribute)):
                    equivalencies = all_flux_unit_conversion_equivs(cube_wave=wave)
                    value = flux_conversion_general(value, unit,
                                                    self.app._get_display_unit(attribute),
                                                    equivalencies,
                                                    with_unit=False)
                    unit = self.app._get_display_unit(attribute)

            elif isinstance(viewer, (CubevizImageView, RampvizImageView)):
                arr = image.get_component(attribute).data
                unit = u.Unit(image.get_component(attribute).units)
                value = self._get_cube_value(
                    image, arr, x, y, viewer
                )

                # We don't want to convert for things like moment maps, so check
                # physical type If unit is flux per pix2, the type will be
                # 'unknown' rather than surface brightness, so multiply out pix2
                # and check if the numerator is a spectral/photon flux density
                if check_if_unit_is_per_solid_angle(unit, return_unit=True) == PIX2:
                    physical_type = (unit * PIX2).physical_type
                else:
                    physical_type = unit.physical_type

                valid_physical_types = ["spectral flux density",
                                        "surface brightness",
                                        "surface brightness wav",
                                        "photon surface brightness wav",
                                        "photon surface brightness",
                                        "power density/spectral flux density wav",
                                        "photon flux density wav",
                                        "photon flux density"]

                if str(physical_type) in valid_physical_types and self.image_unit is not None:

                    # Create list of potentially needed equivalencies for flux/sb unit conversions
                    pixar_sr = self.app.data_collection[0].meta.get('PIXAR_SR', 1)
                    cube_wave = viewer.slice_value * u.Unit(self.app._get_display_unit('spectral'))

                    equivalencies = all_flux_unit_conversion_equivs(pixar_sr,
                                                                    cube_wave)

                    value = flux_conversion_general(value, unit, u.Unit(self.image_unit),
                                                    equivalencies, with_unit=False)
                    unit = self.image_unit

                if associated_dq_layers is not None:
                    associated_dq_layer = associated_dq_layers[0]
                    dq_attribute = associated_dq_layer.state.attribute
                    dq_data = associated_dq_layer.layer.get_data(dq_attribute)
                    dq_value = self._get_cube_value(image, dq_data, x, y, viewer)

            self.row1b_title = 'Value'

            if associated_dq_layers is not None:
                if np.isnan(dq_value):
                    dq_text = ''
                else:
                    dq_text = f' (DQ: {int(dq_value):d})'
            else:
                dq_text = ''
            self.row1b_text = f'{value:+10.5e} {unit}{dq_text}'
            if not isinstance(value, (float, np.floating)):
                self._dict['value'] = float(value)
            else:
                self._dict['value'] = value
            self._dict['value:unit'] = str(unit)
            self._dict['value:unreliable'] = unreliable_pixel
        else:
            self.row1b_title = ''
            self.row1b_text = ''

        if isinstance(viewer, MosvizProfile2DView):
            self.marks[viewer._reference_id].update_xy([x, x], [0, 1])
            self.marks[viewer._reference_id].visible = True

            for matched_marker_id in self._matched_markers.get(viewer._reference_id, []):
                if coords_status and hasattr(getattr(image, 'coords', None), 'pixel_to_world'):
                    # should already have wave computed from setting the coords-info
                    matched_viewer = self.app.get_viewer(matched_marker_id.split(':matched')[0])
                    wave = wave.to_value(matched_viewer.state.x_display_unit)
                    self.marks[matched_marker_id].update_xy([wave, wave], [0, 1])
                    self.marks[matched_marker_id].visible = True
                else:
                    self.marks[matched_marker_id].visible = False

    def _spectrum_viewer_update(self, viewer, x, y):
        def _cursor_fallback():
            self._dict['axes_x'] = x
            self._dict['axes_x:unit'] = str(viewer.state.x_display_unit)
            self._dict['axes_y'] = y
            self._dict['axes_y:unit'] = str(viewer.state.y_display_unit)
            self._dict['data_label'] = ''

        def _copy_axes_to_spectral():
            self._dict['spectral_axis'] = self._dict['axes_x']
            self._dict['spectral_axis:unit'] = self._dict['axes_x:unit']
            self._dict['value'] = self._dict['axes_y']
            self._dict['value:unit'] = self._dict['axes_y:unit']

        if not len(viewer.state.layers):
            return

        self.row1a_title = 'Cursor'
        self.row1a_text = f'{x:10.5e}, {y:10.5e}'

        # show the locked marker/coords only if either no tool or the default tool is active
        if self.dataset.selected == 'none':
            self.row2_title = '\u00A0'
            self.row2_text = ''
            self.row3_title = '\u00A0'
            self.row3_text = ''
            self.icon = 'mdi-cursor-default'
            self.marks[viewer._reference_id].visible = False
            _cursor_fallback()
            _copy_axes_to_spectral()
            return

        # Snap to the closest data point, not the actual mouse location.
        sp = None
        closest_i = None
        closest_wave = None
        closest_flux = None
        closest_icon = 'mdi-cursor-default'
        closest_distance = None
        for lyr in viewer.state.layers:
            if self.dataset.selected == 'auto' and not lyr.visible:
                continue
            if self.dataset.selected != 'auto' and self.dataset.selected != lyr.layer.label:
                continue

            if ((not isinstance(lyr.layer, BaseData)) or (lyr.layer.ndim not in (1, 3))):
                continue
            data_label = lyr.layer.label

            try:
                # Cache should have been populated when spectrum was first plotted.
                # But if not (maybe user changed statistic), we cache it here too.
                cache_key = lyr.layer.label
                if cache_key in self.app._get_object_cache:
                    sp = self.app._get_object_cache[cache_key]
                else:
                    sp = self._specviz_helper.get_data(data_label=data_label)
                    self.app._get_object_cache[cache_key] = sp

                # Calculations have to happen in the frame of viewer display units.
                disp_wave = sp.spectral_axis.to_value(viewer.state.x_display_unit, u.spectral())

                # temporarily here, may be removed after upstream units handling
                # or will be generalized for any sb <-> flux
                # Create list of potentially needed equivalencies for flux/sb unit conversions
                pixar_sr = self.app.data_collection[0].meta.get('PIXAR_SR', 1)
                equivalencies = all_flux_unit_conversion_equivs(pixar_sr,
                                                                sp.spectral_axis)

                if sp.flux.unit is not None and viewer.state.y_display_unit is not None:
                    disp_flux = flux_conversion_general(sp.flux.value,
                                                        sp.flux.unit,
                                                        viewer.state.y_display_unit,
                                                        equivalencies, with_unit=False)  # noqa: E501
                else:
                    disp_flux = sp.flux

                # Out of range in spectral axis.
                if (self.dataset.selected != lyr.layer.label and
                        (x < disp_wave.min() or x > disp_wave.max())):
                    continue

                cur_i = np.argmin(abs(disp_wave - x))
                cur_wave = disp_wave[cur_i]
                cur_flux = disp_flux[cur_i]

                dx = cur_wave - x
                dy = cur_flux - y
                cur_distance = math.sqrt(dx * dx + dy * dy)
                if (closest_distance is None) or (cur_distance < closest_distance):
                    closest_distance = cur_distance
                    closest_i = cur_i
                    closest_wave = cur_wave
                    closest_flux = cur_flux
                    closest_icon = self.app.state.layer_icons.get(lyr.layer.label, '')
                    self._dict['data_label'] = lyr.layer.label
            except Exception:  # nosec
                # Something is loaded but not the right thing
                continue

        if closest_wave is None:
            self.row2_title = '\u00A0'
            self.row2_text = ''
            self.row3_title = '\u00A0'
            self.row3_text = ''
            self.icon = 'mdi-cursor-default'
            self.marks[viewer._reference_id].visible = False
            _cursor_fallback()
            _copy_axes_to_spectral()
            return

        self.row2_title = 'Wave'
        self.row2_text = f'{closest_wave:10.5e} {viewer.state.x_display_unit}'
        self._dict['axes_x'] = closest_wave
        self._dict['axes_x:unit'] = viewer.state.x_display_unit
        if viewer.state.x_display_unit != u.pix:
            self.row2_text += f' ({int(closest_i)} pix)'
            if self.app.config == 'cubeviz':
                # float to be compatible with nan
                self._dict['slice'] = float(closest_i)
                self._dict['spectral_axis'] = closest_wave
                self._dict['spectral_axis:unit'] = viewer.state.x_display_unit
            else:
                # float to be compatible with nan
                self._dict['index'] = float(closest_i)

        if viewer.state.y_display_unit is None:
            flux_unit = ""
        else:
            flux_unit = viewer.state.y_display_unit
        self.row3_title = 'Flux'
        self.row3_text = f'{closest_flux:10.5e} {flux_unit}'
        self._dict['axes_y'] = closest_flux
        self._dict['axes_y:unit'] = str(viewer.state.y_display_unit)

        if closest_icon is not None:
            self.icon = closest_icon
        else:
            self.icon = ""

        self.marks[viewer._reference_id].update_xy([closest_wave], [closest_flux])
        self.marks[viewer._reference_id].visible = True
        for matched_marker_id in self._matched_markers.get(viewer._reference_id, []):
            # NOTE: this currently assumes the the matched marker is a vertical line with a
            # normalized y-scale
            self.marks[matched_marker_id].update_xy([closest_i, closest_i], [0, 1])
            self.marks[matched_marker_id].visible = True
        _copy_axes_to_spectral()



"""
Utility functions for subset plugin these don't belong in
glue.core.region_translators, as they deal with conversions that
only happen within this plugin (e.g between region/'subset defintion')
"""

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from regions import CircleAnnulusSkyRegion, CircleSkyRegion, EllipseSkyRegion, RectangleSkyRegion

from jdaviz.core.region_translators import regions2roi

__all__ = []


def _subset_def_to_region(subset_type, sub, val='value', name='name'):
    """Subset definitions are carried through the plugin in a list of dictionaries,
    one entry for each ROI attribute of that subset along with its value,
    its previous value, and the name to display for that attribute in the UI.
    (see _unpack_get_subsets_for_ui for how these are constructed).

    This function takes one of those dictionary lists describing a subset,
    and converts it to a Regions object.

    Parameters
    ----------
    subset_type : str
        Name of ROI class (CircularROI/TrueCircularROI, EllipticalROI,
        CircularAnnulusROI, or RectangularROI)
    sub : list of dict
        List of dictionaries defining the subset, with one dict for each
        subset attribute (see _unpack_get_subsets_for_ui for how these
        are defined)
    val : str
        Key to get subset attribute (e.g radius, xc, yc) value. Will
        normally be 'value' or 'orig' within plugin to get the new and
        previous values of the attribute, respectivley.
    name : str
        Key to get subset attribute name. Will normally be 'name'.

    Returns
    -------
    reg : `SkyRegion` or `PixelRegion`
        A SkyRegion or PixelRegion object (depending on what
        `region_type` is) of the subset that the dictionary `sub`
        represents.
    """

    if "CircularROI" in subset_type:  # TrueCircular or Circular
        reg = CircleSkyRegion(center=SkyCoord(*[x[val] for x in sub if 'Center' in x[name]]*u.deg),
                              radius=[x[val] for x in sub if 'Radius' in x[name]][0]*u.deg)
    elif subset_type == "EllipticalROI":
        rads = [x[val] for x in sub if 'Radius' in x[name]]
        reg = EllipseSkyRegion(SkyCoord(*[x[val] for x in sub if 'Center' in x[name]]*u.deg),
                               width=2.*rads[0]*u.deg,
                               height=2.*rads[1]*u.deg,
                               angle=[x[val] for x in sub if 'Angle' in x[name]][0]*u.deg)

    elif subset_type == "CircularAnnulusROI":
        rads = [x[val] for x in sub if 'Radius' in x['name']]
        reg = CircleAnnulusSkyRegion(SkyCoord(*[x[val] for x in sub if 'Center' in x[name]]*u.deg),
                                     inner_radius=rads[0]*u.deg,
                                     outer_radius=rads[1]*u.deg)

    elif subset_type == "RectangularROI":
        xmin, ymin = [x[val] for x in sub if 'min' in x[name]]
        xmax, ymax = [x[val] for x in sub if 'max' in x[name]]
        angle = [x[val] for x in sub if 'Angle' in x[name]][0]
        width = xmax - xmin
        height = ymax - ymin
        center = ((xmin+xmax) / 2, (ymin+ymax) / 2)
        reg = RectangleSkyRegion(center=SkyCoord(*center*u.deg),
                                 width=width*u.deg,
                                 height=height*u.deg,
                                 angle=angle*u.deg)

    else:  # pragma: no cover
        raise NotImplementedError(f"Failed to convert {subset_type} to region.")

    return reg


def _sky_region_to_subset_def(sky_region, _around_decimals=6):
    """Generates a 'subset_definition' list of dictionaries from a
    `regions.SkyRegion`object.

    Parameters
    ----------
    sky_region : `regions.SkyRegion`
        Name of ROI class (CircularROI/TrueCircularROI, EllipticalROI,
        CircularAnnulusROI, or RectangularROI).
    _around_decimals : int
        Rounding for 'theta', if present.

    Returns
    -------
    deff : list of dict
        List of dictionaries, each sub-dictionary describing a
        subset attribute.
    """

    if isinstance(sky_region, CircleSkyRegion):
        x = sky_region.center.ra.deg
        y = sky_region.center.dec.deg
        r = sky_region.radius.deg
        deff = [{"name": 'RA Center (degrees)', "att": "xc", "value": x, "orig": x},
                {"name": 'Dec Center (degrees)', "att": "yc", "value": y, "orig": y},
                {"name": 'Radius (degrees)', "att": "radius", "value": r, "orig": r}]

    elif isinstance(sky_region, EllipseSkyRegion):
        xc = sky_region.center.ra.deg
        yc = sky_region.center.dec.deg
        rx = sky_region.width.to_value(u.deg) * 0.5
        ry = sky_region.height.to_value(u.deg) * 0.5
        ang = sky_region.angle.to_value(u.deg)
        theta = np.around(ang, decimals=_around_decimals)
        deff = [{"name": "RA Center (degrees)", "att": "xc", "value": xc, "orig": xc},
                {"name": "Dec Center (degrees)", "att": "yc", "value": yc, "orig": yc},
                {"name": "RA Radius (degrees)", "att": "radius_x", "value": rx, "orig": rx},
                {"name": "Dec Radius (degrees)", "att": "radius_y", "value": ry, "orig": ry},
                {"name": "Angle", "att": "theta", "value": theta, "orig": theta}]

    elif isinstance(sky_region, CircleAnnulusSkyRegion):
        xc = sky_region.center.ra.deg
        yc = sky_region.center.dec.deg
        inner_r = sky_region.inner_radius.to_value(u.deg)
        outer_r = sky_region.outer_radius.to_value(u.deg)
        deff = [{"name": "RA Center (degrees)", "att": "xc", "value": xc, "orig": xc},
                {"name": "Dec Center (degrees)", "att": "yc", "value": yc, "orig": yc},
                {"name": "Inner Radius (degrees)", "att": "inner_radius",
                 "value": inner_r, "orig": inner_r},
                {"name": "Outer Radius (degrees)", "att": "outer_radius",
                 "value": outer_r, "orig": outer_r}]

    elif isinstance(sky_region, RectangleSkyRegion):
        deff = []
        dsky_ra = (sky_region.width).to_value(u.deg) * 0.5
        dsky_dec = (sky_region.height).to_value(u.deg) * 0.5
        mins_maxs = {'xmin': sky_region.center.ra.deg - dsky_ra,
                     'xmax': sky_region.center.ra.deg + dsky_ra,
                     'ymin': sky_region.center.dec.deg - dsky_dec,
                     'ymax': sky_region.center.dec.deg + dsky_dec}
        for att in ("Xmin", "Xmax", "Ymin", "Ymax"):
            val = mins_maxs[att.lower()]
            name = att.replace('X', 'RA ').replace('Y', 'Dec ')
            deff.append({"name": f"{name} (degrees)", "att": att.lower(),
                         "value": val, "orig": val})
        theta = sky_region.angle.to_value(u.deg)
        deff.append({"name": "Angle", "att": "theta", "value": theta, "orig": theta})

    else:  # pragma: no cover
        raise NotImplementedError(f"Failed to convert {sky_region} to subset definition.")

    return deff


def _get_pixregion_params_in_dict(region):
    """Given a Region, returns a dictionary with the attribute names
    of the corresponding ROI type (i.e., CirclePixelRegion and CircularROI).

    This makes use of `jdaviz.core.region_translators.regions2roi`, but instead
    of returning an ROI object it returns a dictionary of parameters that
    would be used to create that same ROI object. This involves changing some
    keys and deleting some, which is why that `regions2roi` can't be
    called directly.
    """

    roi = regions2roi(region)
    region_dict = roi.__dict__.copy()

    if 'center' in region_dict.keys():
        region_dict['xc'] = region_dict['center'].x
        region_dict['yc'] = region_dict['center'].y
        del region_dict['center']

    region_dict.pop('meta', None)
    region_dict.pop('visual', None)

    return region_dict



import numpy as np
import specutils
from astropy import units as u
from astropy.nddata import CCDData
from astropy.utils import minversion
from traitlets import Bool, List, Unicode, observe
from specutils import manipulation, analysis, Spectrum1D

from jdaviz.core.custom_traitlets import IntHandleEmpty, FloatHandleEmpty
from jdaviz.core.events import SnackbarMessage, GlobalDisplayUnitChanged
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        DatasetSelect, DatasetSelectMixin,
                                        SpectralSubsetSelectMixin,
                                        AddResultsMixin,
                                        SelectPluginComponent,
                                        SpectralContinuumMixin,
                                        skip_if_no_updates_since_last_active,
                                        with_spinner)
from jdaviz.core.unit_conversion_utils import convert_integrated_sb_unit
from jdaviz.core.user_api import PluginUserApi

__all__ = ['MomentMap']

SPECUTILS_LT_1_15_1 = not minversion(specutils, "1.15.1.dev")

moment_unit_options = {0: ["Surface Brightness"],
                       1: ["Velocity", "Spectral Unit"],
                       2: ["Velocity", "Velocity^N"]}


@tray_registry('cubeviz-moment-maps', label="Moment Maps",
               viewer_requirements=['spectrum', 'image'])
class MomentMap(PluginTemplateMixin, DatasetSelectMixin, SpectralSubsetSelectMixin,
                SpectralContinuumMixin, AddResultsMixin):
    """
    See the :ref:`Moment Maps Plugin Documentation <moment-maps>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset to use for computing line statistics.
    * ``spectral_subset`` (:class:`~jdaviz.core.template_mixin.SubsetSelect`):
      Subset to use for the line, or ``Entire Spectrum``.
    * ``continuum`` (:class:`~jdaviz.core.template_mixin.SubsetSelect`):
      Subset to use for the continuum, or ``None`` to skip continuum subtraction,
      or ``Surrounding`` to use a region surrounding the
      subset set in ``spectral_subset``.
    * ``continuum_dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset of the extracted 1D spectrum to use when visualizing the continuum.
      The continuum will be redetermined based on the input cube (``dataset``) when
      computing the moment map.
    * ``continuum_width``:
      Width, relative to the overall line spectral region, to fit the linear continuum
      (excluding the region containing the line). If 1, will use endpoints within line region
      only.
    * ``n_moment``
    * ``output_unit``
      Choice of "Wavelength" or "Velocity", applicable for n_moment >= 1.
    * ``reference_wavelength``
      Reference wavelength for conversion of output to velocity units.
    * ``add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * :meth:`calculate_moment`
    """
    template_file = __file__, "moment_maps.vue"
    uses_active_status = Bool(True).tag(sync=True)

    continuum_dataset_items = List().tag(sync=True)
    continuum_dataset_selected = Unicode().tag(sync=True)

    n_moment = IntHandleEmpty(0).tag(sync=True)
    filename = Unicode().tag(sync=True)
    moment_available = Bool(False).tag(sync=True)
    output_unit_items = List().tag(sync=True)
    output_radio_items = List().tag(sync=True)
    output_unit_selected = Unicode().tag(sync=True)
    reference_wavelength = FloatHandleEmpty().tag(sync=True)
    dataset_spectral_unit = Unicode().tag(sync=True)

    # export_enabled controls whether saving the moment map to a file is enabled via the UI.  This
    # is a temporary measure to allow server-installations to disable saving server-side until
    # saving client-side is supported
    export_enabled = Bool(True).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Create a 2D image from a data cube.'

        self.moment = None

        self.continuum_dataset = DatasetSelect(self,
                                               'continuum_dataset_items',
                                               'continuum_dataset_selected',
                                               filters=['not_child_layer',
                                                        'layer_in_spectrum_viewer'])
        # since the continuum is just an approximation preview,
        # automatically convert the units instead of recomputing
        self.continuum_auto_update_units = True

        # when plugin is initialized, there won't be a dataset selected, so
        # call the output unit 'Flux' for now (rather than surface brightness).
        # the appropriate label will be chosen once a dataset is selected, and/or
        # unit conversion is done
        self.output_unit = SelectPluginComponent(self,
                                                 items='output_unit_items',
                                                 selected='output_unit_selected',
                                                 manual_options=['Surface Brightness',
                                                                 'Spectral Unit',
                                                                 'Velocity',
                                                                 'Velocity^N'])

        self.dataset.add_filter('is_cube')
        self.add_results.viewer.filters = ['is_image_viewer']
        self.hub.subscribe(self, GlobalDisplayUnitChanged,
                           handler=self._set_data_units)

        if self.app.state.settings.get('server_is_remote', False):
            # when the server is remote, saving the file in python would save on the server, not
            # on the user's machine, so export support in cubeviz should be disabled
            self.export_enabled = False

    @property
    def _default_image_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_spectrum_viewer_reference_name', 'flux-viewer'
        )

    @property
    def _default_spectrum_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_spectrum_viewer_reference_name', 'spectrum-viewer'
        )

    @property
    def user_api(self):
        # NOTE: leaving save_as_fits out for now - we may want a more general API to do that
        # accross all plugins at some point
        return PluginUserApi(self, expose=('dataset', 'spectral_subset',
                                           'continuum', 'continuum_dataset', 'continuum_width',
                                           'n_moment',
                                           'output_unit', 'reference_wavelength',
                                           'add_results', 'calculate_moment'))

    @property
    def slice_display_unit_name(self):
        return 'spectral'

    @observe('is_active')
    def _is_active_changed(self, msg):
        for pos, mark in self.continuum_marks.items():
            mark.visible = self.is_active
        self._calculate_continuum(msg)

    @observe("dataset_selected", "dataset_items", "n_moment")
    def _set_default_results_label(self, event={}):
        label_comps = []
        if hasattr(self, 'dataset') and len(self.dataset.labels) > 1:
            label_comps += [self.dataset_selected]
        label_comps += [f"moment {self.n_moment}"]
        self.results_label_default = " ".join(label_comps)

    @observe("dataset_selected", "n_moment")
    def _set_data_units(self, event={}):

        if isinstance(self.n_moment, str) or self.n_moment < 0:
            return
        unit_options_index = min(self.n_moment, 2)
        if self.output_unit_selected not in moment_unit_options[unit_options_index]:
            self.output_unit_selected = moment_unit_options[unit_options_index][0]
        self.send_state("output_unit_selected")

        unit_dict = {"Surface Brightness": "",
                     "Spectral Unit": "",
                     "Velocity": "km/s",
                     "Velocity^N": f"km{self.n_moment}/s{self.n_moment}"}

        if self.dataset_selected != "":
            # Spectral axis is first in this list
            data = self.app.data_collection[self.dataset_selected]
            if (self.spectrum_viewer and hasattr(self.spectrum_viewer.state, 'x_display_unit')
                    and self.spectrum_viewer.state.x_display_unit is not None):
                sunit = self.spectrum_viewer.state.x_display_unit
            elif self.app.data_collection[self.dataset_selected].coords is not None:
                sunit = data.coords.world_axis_units[0]
            else:
                sunit = ""
            self.dataset_spectral_unit = sunit
            unit_dict["Spectral Unit"] = sunit
            unit_dict["Surface Brightness"] = str(self.moment_zero_unit)

        # Update units in selection item dictionary
        for item in self.output_unit_items:
            item["unit_str"] = unit_dict[item["label"]]

        # Filter what we want based on n_moment
        if self.n_moment == 0:
            self.output_radio_items = [self.output_unit_items[0],]
        elif self.n_moment == 1:
            self.output_radio_items = self.output_unit_items[1:3]
        else:
            self.output_radio_items = self.output_unit_items[2:]

        # Force Traitlets to update
        self.send_state("output_unit_items")
        self.send_state("output_radio_items")

    @observe("dataset_selected", "spectral_subset_selected",
             "continuum_subset_selected", "continuum_dataset_selected", "continuum_width")
    @skip_if_no_updates_since_last_active()
    def _calculate_continuum(self, msg=None):
        if not hasattr(self, 'dataset') or self.app._jdaviz_helper is None:  # noqa
            # during initial init, this can trigger before the component is initialized
            return
        if self.continuum_dataset_selected == '':
            # NOTE: we could send self.dataset.selected through
            # spectral_extraction._extract_in_new_instance() to get a spectrum
            # for the selected/default cube,
            # but there is no visible spectrum to even show under the continuum
            return
        # NOTE: there is no use in caching this, as the continuum will need to be re-computed
        # per-spaxel to use within calculating the moment map
        _ = self._get_continuum(self.continuum_dataset,
                                self.spectral_subset,
                                update_marks=True)

    @with_spinner()
    def calculate_moment(self, add_data=True):
        """
        Calculate the moment map

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting data object to the app according to ``add_results``.
        """

        # Check to make sure API use hasn't put us into an invalid state.

        try:
            n_moment = int(self.n_moment)
            if n_moment < 0:
                raise ValueError("Moment must be a positive integer")
        except ValueError:
            raise ValueError("Moment must be a positive integer")

        unit_options_index = 2 if n_moment > 2 else n_moment
        if self.output_unit_selected not in moment_unit_options[unit_options_index]:
            raise ValueError("Selected output units must be in "
                             f"{moment_unit_options[unit_options_index]} for "
                             f"moment {self.n_moment}")

        if self.continuum.selected == 'None':
            if "_orig_spec" in self.dataset.selected_obj.meta:
                cube = self.dataset.selected_obj.meta["_orig_spec"]
            else:
                cube = self.dataset.selected_obj
        else:
            _, _, cube = self._get_continuum(self.continuum_dataset,
                                             self.spectral_subset,
                                             update_marks=False,
                                             per_pixel=True)

        # slice out desired region
        # TODO: should we add a warning for a composite spectral subset?
        if self.spectral_subset.selected == "Entire Spectrum":
            spec_reg = None
        else:
            spec_reg = self.app.get_subsets(self.spectral_subset.selected,
                                            simplify_spectral=True,
                                            use_display_units=True)
        # We need to convert the spectral region to the display units

        if spec_reg is None:
            slab = cube
        else:
            slab = manipulation.extract_region(cube, spec_reg)

        # Calculate the moment and convert to CCDData to add to the viewers
        # Need transpose to align JWST mirror shape: This is because specutils
        # arrange the array shape to be (nx, ny, nz) but 2D visualization
        # assumes (ny, nx) as per row-major convention.

        # Extract 2D WCS from input cube.
        data = self.dataset.selected_dc_item
        # Similar to coords_info logic.
        if '_orig_spec' in getattr(data, 'meta', {}):
            w = data.meta['_orig_spec'].wcs
        else:
            w = data.coords
        data_wcs = getattr(w, 'celestial', None)
        if data_wcs:
            data_wcs = data_wcs.swapaxes(0, 1)  # We also transpose WCS to match.

        # Convert spectral axis to velocity units if desired output is in velocity
        if n_moment > 0 and self.output_unit_selected.lower().startswith("velocity"):
            # Catch this if called from API
            if not self.reference_wavelength > 0.0:
                raise ValueError("reference_wavelength must be set for output in velocity units.")

            ref_wavelength = self.reference_wavelength * u.Unit(self.dataset_spectral_unit)
            slab_sa = slab.spectral_axis.to("km/s", doppler_convention="relativistic",
                                            doppler_rest=ref_wavelength)
            slab = Spectrum1D(slab.flux, slab_sa, uncertainty=slab.uncertainty)
        # Otherwise convert spectral axis to display units, have to do frequency <-> wavelength
        # before calculating
        else:
            slab_sa = slab.spectral_axis.to(self.app._get_display_unit('spectral'))
            slab = Spectrum1D(slab.flux, slab_sa, uncertainty=slab.uncertainty)

        # Finally actually calculate the moment
        self.moment = analysis.moment(slab, order=n_moment).T
        # If n>1 and velocity is desired, need to take nth root of result
        if n_moment > 0 and self.output_unit_selected.lower() == "velocity":
            self.moment = np.power(self.moment, 1/self.n_moment)

        # convert units for moment 0, which is the only currently supported
        # moment for using converted units.
        if n_moment == 0:
            if self.moment_zero_unit != self.moment.unit:
                spectral_axis_unit = u.Unit(self.spectrum_viewer.state.x_display_unit)

                # if the flux unit is a per-frequency unit but the spectral axis unit
                # is a wavelength, or vice versa, we need to convert the spectral axis
                # unit that the flux was integrated over so they are compatible for
                # unit conversion (e.g., Jy m / sr needs to become Jy Hz / sr, and
                # (erg Hz)/(s * cm**2 * AA) needs to become (erg)/(s * cm**2)
                desired_freq_unit = spectral_axis_unit if spectral_axis_unit.physical_type == 'frequency' else u.Hz  # noqa E501
                desired_length_unit = spectral_axis_unit if spectral_axis_unit.physical_type == 'length' else u.AA  # noqa E501
                moment_temp = convert_integrated_sb_unit(self.moment,
                                                         spectral_axis_unit,
                                                         desired_freq_unit,
                                                         desired_length_unit)
                moment_zero_unit_temp = convert_integrated_sb_unit(1 * u.Unit(self.moment_zero_unit),  # noqa E501
                                                                   spectral_axis_unit,
                                                                   desired_freq_unit,
                                                                   desired_length_unit)

                moment = moment_temp.to(moment_zero_unit_temp.unit, u.spectral())

                # if flux and spectral axis units were incompatible in terms of freq/wav
                # and needed to be converted to an intermediate unit for conversion, then
                # re-instate the original chosen units (e.g Jy m /sr was converted to Jy Hz / sr
                # for unit conversion, now back to Jy m / sr)
                if spectral_axis_unit not in moment.unit.bases:
                    if spectral_axis_unit.physical_type == 'frequency':
                        moment *= (1*desired_length_unit).to(desired_freq_unit,
                                                             u.spectral()) / desired_length_unit
                    elif spectral_axis_unit.physical_type == 'length':
                        moment *= (1*desired_freq_unit).to(desired_length_unit,
                                                           u.spectral()) / desired_freq_unit

                self.moment = moment

        # Reattach the WCS so we can load the result
        self.moment = CCDData(self.moment, wcs=data_wcs)

        fname_label = self.dataset_selected.replace("[", "_").replace("]", "")
        self.filename = f"moment{n_moment}_{fname_label}.fits"

        if add_data:
            self.add_results.add_results_from_plugin(self.moment)

            msg = SnackbarMessage("{} added to data collection".format(self.results_label),
                                  sender=self, color="success")
            self.hub.broadcast(msg)

        self.moment_available = True

        return self.moment

    @property
    def moment_zero_unit(self):
        if self.spectrum_viewer.state.x_display_unit is not None:
            return (
                u.Unit(self.app._get_display_unit('sb')) *
                u.Unit(self.spectrum_viewer.state.x_display_unit)
            )
        return u.dimensionless_unscaled

    @property
    def spectral_unit_selected(self):
        return self.app._get_display_unit('spectral')

    def vue_calculate_moment(self, *args):
        self.calculate_moment(add_data=True)

    def vue_overwrite_fits(self, *args):
        """Attempt to force writing the moment map if the user confirms the desire to overwrite."""
        self.overwrite_warn = False



from astropy.tests.helper import assert_quantity_allclose
import astropy.units as u
from glue.viewers.profile.state import FUNCTIONS as COLLAPSE_FUNCTIONS
import numpy as np
import pytest
from specutils import Spectrum1D

from jdaviz import Cubeviz

import warnings
warnings.filterwarnings('ignore')

# Maps input spectrum flux unit to expected line analysis flux unit
expected_lineflux_results = {
    u.Jy/u.sr: u.Unit('W/(m2)'),
    u.Jy: u.Unit('W/m2'),
    u.Unit('W/m2/m')/u.sr: u.Unit('W/(m2)'),
    u.Unit('W/m2/m'): u.Unit('W/m2')
}

test_cases = list(expected_lineflux_results.keys())


# Test cases for unit gaussian tests
def _gauss_with_unity_area(x, mean, sigma):
    ''' Gaussian function with area = 1 unit '''
    dx = x - mean
    n = 1/(sigma*np.sqrt(2*np.pi))
    g = n * np.exp(-dx**2/(2*sigma**2))
    return g


mn = 1.5
sig = 0.01

# 5 cases: replace fl_wave with fnu_freq, fnu_wave, flam_wave, flam_freq, or fl_wave
# to try the different cases here
unit_flux_gaussian_test_cases = []

# unit-flux gaussian in frequency space
freq = np.arange(1, 2, 0.001)*u.Hz
flux_freq = _gauss_with_unity_area(freq.value, mn, sig)*1.0E26*u.Jy
fnu_freq = Spectrum1D(spectral_axis=freq, flux=flux_freq)
unit_flux_gaussian_test_cases.append(fnu_freq)
fnu_wave = Spectrum1D(spectral_axis=fnu_freq.wavelength, flux=flux_freq)
unit_flux_gaussian_test_cases.append(fnu_wave)

# unit-flux gaussian in wavelength space
lam = np.arange(1, 2, 0.001)*u.m
flux_wave = _gauss_with_unity_area(lam.value, mn, sig)*1.0*u.W/u.m**2/u.m
flam_wave = Spectrum1D(spectral_axis=lam, flux=flux_wave)
unit_flux_gaussian_test_cases.append(flam_wave)
flam_freq = Spectrum1D(spectral_axis=flam_wave.frequency, flux=flux_wave)
unit_flux_gaussian_test_cases.append(flam_freq)


def _calculate_line_flux(viz_helper):
    '''
    Returns the line flux calculation by opening the Line Analysis Plugin
    Assumes the plugin hasn't been opened yet
    '''
    # Open the plugin and force the calculation
    viz_helper.app.state.drawer_content = 'plugins'
    line_analysis_plugin = viz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    line_analysis_plugin.keep_active = True

    # Retrieve Results
    for result in line_analysis_plugin.results:
        if result['function'] == 'Line Flux':
            return result


@pytest.mark.filterwarnings(r"ignore:.* contains multiple slashes")
@pytest.mark.filterwarnings(r"ignore:.* apply_slider_redshift")
@pytest.mark.parametrize('spectra_fluxunit', test_cases)
def test_cubeviz_collapse_fluxunits(spectrum1d_cube_custom_fluxunit, spectra_fluxunit):
    ''' Calculates line flux and checks the units for each collapse function '''
    data = spectrum1d_cube_custom_fluxunit(spectra_fluxunit)
    for function in COLLAPSE_FUNCTIONS:
        # Initialize Cubeviz with specific data and collapse function
        cubeviz_helper = Cubeviz()
        data_label = "Test Cube"
        cubeviz_helper.load_data(data, data_label=data_label)
        cubeviz_helper.app.get_viewer('spectrum-viewer').state.function = function

        lineflux_result = _calculate_line_flux(cubeviz_helper)
        autocollapsed_spectrum_unit = (cubeviz_helper.
                                       specviz.get_spectra()["Spectrum (sum)"].flux.unit)
        # Futureproofing: Eventually Cubeviz autocollapse will change the flux units of the
        # spectra depending on whether the spectrum was collapsed OVER SPAXELS or not. Only
        # do the assertion check if we KNOW what the expected lineflux results should be
        if autocollapsed_spectrum_unit in expected_lineflux_results.keys():
            assert u.Unit(lineflux_result['unit']) == expected_lineflux_results[spectra_fluxunit]


@pytest.mark.filterwarnings(r"ignore:.* contains multiple slashes")
@pytest.mark.parametrize('test_case', unit_flux_gaussian_test_cases)
def test_unit_gaussian(specviz_helper, test_case):
    '''
    Test an Area 1 Gaussian and ensure the result returns in W/m2
    Test provided by Patrick Ogle
    '''
    specviz_helper.load_data(test_case)

    lineflux_result = _calculate_line_flux(specviz_helper)
    assert_quantity_allclose(float(lineflux_result['result']) * u.Unit(lineflux_result['unit']),
                             1*u.Unit('W/m2'))


@pytest.mark.filterwarnings(r"ignore:.* contains multiple slashes")
def test_unit_gaussian_mixed_units_per_steradian(specviz_helper):
    '''
    A special unit test of Area 1 with mixed units. Should return W/m2sr
    Test provided by Patrick Ogle
    '''
    # unit-flux gaussian in wavelength space, mixed units, per steradian
    lam_a = np.arange(1, 2, 0.001)*u.Angstrom
    # test changed from Surface Brightness to Flux,
    # u.erg/u.s/u.cm**2/u.Angstrom/u.sr in untranslatable units (check unit_conversion.py)
    flx_wave = _gauss_with_unity_area(lam_a.value, mn, sig)*1E3*u.erg/u.s/u.cm**2/u.Angstrom
    fl_wave = Spectrum1D(spectral_axis=lam_a, flux=flx_wave)

    specviz_helper.load_data(fl_wave)
    lineflux_result = _calculate_line_flux(specviz_helper)
    assert_quantity_allclose(float(lineflux_result['result']) * u.Unit(lineflux_result['unit']),
                             1*u.Unit('W/(m2)'))



import numpy as np
import astropy.units as u
from astropy.nddata import NDDataArray

from functools import cached_property
from traitlets import Bool, Float, List, Unicode, observe, Int
from glue.core.message import (
    DataCollectionAddMessage, SubsetCreateMessage, SubsetDeleteMessage, SubsetUpdateMessage
)

from jdaviz.core.events import SnackbarMessage, SliceValueUpdatedMessage
from jdaviz.core.marks import PluginLine
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        DatasetSelectMixin,
                                        SelectPluginComponent,
                                        ApertureSubsetSelectMixin,
                                        ApertureSubsetSelect,
                                        AddResultsMixin,
                                        skip_if_not_tray_instance,
                                        skip_if_no_updates_since_last_active,
                                        with_spinner, with_temp_disable)
from jdaviz.core.user_api import PluginUserApi
from jdaviz.configs.cubeviz.plugins.viewers import WithSliceIndicator


__all__ = ['RampExtraction']

rng = np.random.default_rng(seed=42)


@tray_registry(
    'ramp-extraction', label="Ramp Extraction", viewer_requirements='profile'
)
class RampExtraction(PluginTemplateMixin, ApertureSubsetSelectMixin,
                     DatasetSelectMixin, AddResultsMixin):
    """
    See the :ref:`Ramp Extraction Plugin Documentation <ramp-extraction>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``aperture`` (:class:`~jdaviz.core.template_mixin.ApertureSubsetSelect`):
      Subset to use for the ramp extraction, or ``Entire Cube``.
    * ``aperture_method`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      Method to use for extracting a ramp profile
    * ``add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * :meth:`extract`
    """
    template_file = __file__, "ramp_extraction.vue"
    uses_active_status = Bool(True).tag(sync=True)
    show_live_preview = Bool(False).tag(sync=True)
    show_subset_preview = Bool(True).tag(sync=True)
    subset_preview_warning = Bool(False).tag(sync=True)
    subset_preview_limit = Int(250).tag(sync=True)

    active_step = Unicode().tag(sync=True)

    resulting_product_name = Unicode("ramp").tag(sync=True)
    do_auto_extraction = True

    slice_group_value = Float().tag(sync=True)

    function_items = List().tag(sync=True)
    function_selected = Unicode('Mean').tag(sync=True)
    filename = Unicode().tag(sync=True)
    extraction_available = Bool(False).tag(sync=True)
    overwrite_warn = Bool(False).tag(sync=True)

    aperture_method_items = List().tag(sync=True)
    aperture_method_selected = Unicode('Center').tag(sync=True)

    conflicting_aperture_and_function = Bool(False).tag(sync=True)
    conflicting_aperture_error_message = Unicode('Aperture method Exact cannot be selected along'
                                                 ' with Min or Max.').tag(sync=True)

    # export_enabled controls whether saving to a file is enabled via the UI.  This
    # is a temporary measure to allow server-installations to disable saving server-side until
    # saving client-side is supported
    export_enabled = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Extract a ramp from a ramp cube.'

        self.dataset.filters = ['is_flux_cube']

        # TODO: in the future this could be generalized with support in SelectPluginComponent
        self.aperture._default_text = 'Entire Cube'
        self.aperture._manual_options = ['Entire Cube']
        self.aperture.items = [{"label": "Entire Cube"}]
        self.aperture._subset_selected_changed_callback = self._update_extract
        # need to reinitialize choices since we overwrote items and some subsets may already
        # exist.
        self.aperture._initialize_choices()
        self.aperture.select_default()

        self.extracted_ramp = None

        self.function = SelectPluginComponent(
            self,
            items='function_items',
            selected='function_selected',
            manual_options=['Mean', 'Median', 'Min', 'Max', 'Sum']
        )
        self._set_default_results_label()
        self.add_results.viewer.filters = ['is_slice_indicator_viewer']

        self.session.hub.subscribe(self, SliceValueUpdatedMessage,
                                   handler=self._on_slice_changed)

        self.session.hub.subscribe(self, DataCollectionAddMessage,
                                   handler=self._on_data_added)

        self.session.hub.subscribe(self, SubsetCreateMessage,
                                   handler=self._on_subset_update)

        self.session.hub.subscribe(self, SubsetUpdateMessage,
                                   handler=self._on_subset_update)

        self.session.hub.subscribe(self, SubsetDeleteMessage,
                                   handler=self._on_subset_delete)

        self._update_disabled_msg()

        if self.app.state.settings.get('server_is_remote', False):
            # when the server is remote, saving the file in python would save on the server, not
            # on the user's machine, so export support in cubeviz should be disabled
            self.export_enabled = False

    @property
    def integration_viewer(self):
        viewer = self.app.get_viewer(
            self.app._jdaviz_helper._default_integration_viewer_reference_name
        )
        return viewer

    def _on_data_added(self, msg={}):
        # only perform the default collapse after the first data load:
        if len(self.app._jdaviz_helper.cube_cache) and not self.extraction_available:
            self.extract(add_data=True)
            self.integration_viewer._initialize_x_axis()

    def _on_subset_update(self, msg={}):
        if not hasattr(self.app._jdaviz_helper, 'cube_cache'):
            # if called before fully initialized
            return

        subset_lbl = msg.subset.label
        color = msg.subset.style.color
        subset = self.app.get_subsets(subset_lbl)[0]
        region = subset['region']

        if region is None:
            return

        # glue region has transposed coords relative to cached cube:
        region_mask = region.to_mask().to_image(
            self.cube.shape[:-1][::-1]
        ).astype(bool).T
        cube_subset = self.cube[region_mask]  # shape: (N pixels extracted, M groups)

        n_pixels_in_extraction = cube_subset.shape[0]

        if n_pixels_in_extraction < self.subset_preview_limit:
            self.subset_preview_warning = False
            select_from_cube_subset = Ellipsis
        else:
            self.subset_preview_warning = True
            select_from_cube_subset = rng.integers(
                0, n_pixels_in_extraction, size=self.subset_preview_limit
            )

        marks = [
            PluginLine(
                self.integration_viewer, x=np.arange(cube_subset.shape[1]), y=y,
                stroke_width=1, colors=[color], opacities=[0.25], label=subset_lbl,
                visible=self._subset_preview_visible and subset_lbl == self.aperture.selected
            )
            for y in cube_subset[select_from_cube_subset]
        ]

        self.integration_viewer.figure.marks = [
            mark for mark in self.integration_viewer.figure.marks
            if getattr(mark, 'label', None) != subset_lbl
        ] + marks

        self.integration_viewer.reset_limits()

    def _on_subset_delete(self, msg={}):
        subset_lbl = msg.subset.label
        self.integration_viewer.figure.marks = [
            mark for mark in self.integration_viewer.figure.marks
            if getattr(mark, 'label', None) != subset_lbl
        ]

    @observe('is_active', 'show_subset_preview', 'aperture_selected')
    def _update_subset_previews(self, msg={}):
        # remove preview marks for non-selected subsets

        if not hasattr(self.app._jdaviz_helper, '_default_integration_viewer_reference_name'):
            return

        for mark in self.integration_viewer.figure.marks:
            if isinstance(mark, PluginLine) and mark.label is not None:
                new_visibility = (
                    self._subset_preview_visible and
                    self.aperture.selected == mark.label
                )
                if mark.visible != new_visibility:
                    mark.visible = new_visibility

        self.integration_viewer.reset_limits()

    @property
    def _subset_preview_visible(self):
        return self.show_subset_preview and self.is_active

    @property
    def user_api(self):
        expose = [
            'dataset', 'function', 'aperture', 'add_results', 'extract'
        ]

        return PluginUserApi(self, expose=expose)

    @observe('dataset_items')
    def _update_disabled_msg(self, msg={}):
        for data in self.app.data_collection:
            if data.data.ndim == 3:
                self.disabled_msg = ''
                break
        else:
            # no cube-like data loaded.  Once loaded, the parser will unset this
            self.disabled_msg = (
                f"{self.__class__.__name__} requires a 3d cube dataset to be loaded, "
                "please load data to enable this plugin."
            )

    @property
    def live_update_subscriptions(self):
        return {'data': ('dataset',), 'subset': ('aperture', )}

    def __call__(self, add_data=True):
        return self.extract(add_data=add_data)

    @property
    def slice_display_unit_name(self):
        return 'temporal'

    @property
    def spatial_axes(self):
        # Collapse an e.g. 3D ramp cube to 1D ramp profile, assuming that last axis
        # is always the group/resultant index
        return (0, 1)

    @property
    def slice_indicator_viewers(self):
        return [v for v in self.app._viewer_store.values() if isinstance(v, WithSliceIndicator)]

    @observe('active_step', 'is_active')
    def _active_step_changed(self, *args):
        self.aperture._set_mark_visiblities(self.active_step in ('', 'ap', 'extract'))

    @property
    def slice_plugin(self):
        return self.app._jdaviz_helper.plugins['Slice']

    def _extract_in_new_instance(self, dataset=None, function='Mean', subset_lbl=None,
                                 auto_update=False, add_data=False):
        # create a new instance of the Ramp Extraction plugin (to not affect the instance in
        # the tray) and extract the entire cube with defaults.
        plg = self.new()
        plg.dataset.selected = self.dataset.selected
        if subset_lbl is not None:
            plg.aperture.selected = subset_lbl
        plg.function.selected = function
        plg.add_results.auto_update_result = auto_update
        # all other settings remain at their plugin defaults
        return plg(add_data=add_data)

    def _on_slice_changed(self, msg):
        self.slice_group_value = msg.value

    @observe('function_selected', 'aperture_method_selected')
    def _update_aperture_method_on_function_change(self, *args):
        if (self.function_selected.lower() in ('min', 'max') and
                self.aperture_method_selected.lower() != 'center'):
            self.conflicting_aperture_and_function = True
        else:
            self.conflicting_aperture_and_function = False

    @property
    def cube(self):
        return self.app._jdaviz_helper.cube_cache.get(self.dataset.selected)

    @property
    def slice_display_unit(self):
        x_display_unit = self.app._get_display_unit(self.slice_display_unit_name)
        if x_display_unit not in ['None', None]:
            return u.Unit(x_display_unit)
        return u.dimensionless_unscaled

    @property
    def aperture_weight_mask(self):
        cube_shape = self.cube.shape
        if self.aperture.selected != self.aperture.default_text:

            # note: glue subset mask is transposed relative to cube
            region_mask = self.app.get_subsets(
                subset_name=self.aperture.selected
            )[0]['region'].to_mask().to_image(
                cube_shape[:-1]
            ).astype(bool).T
            return region_mask

        return np.ones(cube_shape[:-1]).astype(bool)

    def _extract_from_aperture(self, **kwargs):
        # This plugin collapses over the *spatial axes* (optionally over a spatial subset,
        # defaults to ``No Subset``). Since the Cubeviz parser puts the fluxes
        # and uncertainties in different glue Data objects, we translate the ramp
        # cube and its uncertainties into separate NDDataArrays, then combine them:
        selected_func = self.function_selected.lower()

        if not isinstance(self.aperture, ApertureSubsetSelect):
            raise ValueError("aperture must be an ApertureSubsetSelect object")

        nddata = self.cube
        mask = self.aperture_weight_mask

        if nddata.mask is not None:
            mask = mask & ~nddata.mask

        collapsed = getattr(np, selected_func)(
            nddata.data[mask],
            # after the fancy indexing above, axis=1 corresponds to groups, and
            # operations over axis=0 corresponds to individual pixels:
            axis=0
        )
        if nddata.unit is not None:
            collapsed <<= nddata.unit

        def expand(x):
            # put the resulting 1D profile (counts vs. groups) into the
            # third dimension, which is the group dimension in the
            # original 3D cube:
            return np.expand_dims(x, axis=(0, 1))

        return NDDataArray(
            data=expand(collapsed),
            meta=nddata.meta
        )

    def _preview_x_from_extracted(self, extracted):
        return np.arange(extracted.shape[-1])

    def _preview_y_from_extracted(self, extracted):
        return extracted.data

    @with_spinner()
    def extract(self, add_data=True, **kwargs):
        """
        Extract the ramp profile from the data cube according to the plugin inputs.

        Parameters
        ----------
        add_data : bool, optional
            Whether to load the resulting data back into the application according to
            ``add_results``.
        kwargs : dict
            Additional keyword arguments passed to the NDDataArray collapse operation.
            Examples include ``propagate_uncertainties`` and ``operation_ignores_mask``.
        """
        if self.conflicting_aperture_and_function:
            raise ValueError(self.conflicting_aperture_error_message)

        selected_func = self.function_selected.lower()
        ndd = self._extract_from_aperture(**kwargs)
        self.extracted_ramp = ndd
        self.extraction_available = True
        fname_label = self.dataset_selected.replace("[", "_").replace("]", "")
        self.filename = f"extracted_{selected_func}_{fname_label}.fits"

        if add_data:
            if default_color := self.aperture.selected_item.get('color', None):
                ndd.meta['_default_color'] = default_color
            self.add_results.add_results_from_plugin(ndd)

            snackbar_message = SnackbarMessage(
                f"{self.resulting_product_name.title()} extracted successfully.",
                color="success",
                sender=self)
            self.hub.broadcast(snackbar_message)

        return ndd

    def vue_ramp_extraction(self, *args, **kwargs):
        try:
            self.extract(add_data=True)
        except Exception as e:
            self.hub.broadcast(SnackbarMessage(
                f"Extraction failed: {repr(e)}",
                sender=self, color="error"))

    @observe('aperture_selected', 'function_selected')
    def _set_default_results_label(self, event={}):
        if not hasattr(self, 'aperture'):
            return
        if self.aperture.selected == self.aperture.default_text:
            self.results_label_default = (f"{self.resulting_product_name.title()} "
                                          f"({self.function_selected.lower()})")

        else:
            self.results_label_default = (f"{self.resulting_product_name.title()} "
                                          f"({self.aperture_selected}, "
                                          f"{self.function_selected.lower()})")

    @cached_property
    def marks(self):
        if not self._tray_instance:
            return {}
        # TODO: iterate over self.slice_indicator_viewers and handle adding/removing viewers

        sv = self.slice_indicator_viewers[0]
        marks = {'extract': PluginLine(sv, visible=self.is_active)}
        sv.figure.marks = sv.figure.marks + [marks['extract'],]
        return marks

    def _clear_marks(self):
        for mark in self.marks.values():
            if getattr(mark, 'visible', False):
                mark.visible = False

    @observe('is_active', 'show_live_preview',
             'dataset_selected', 'aperture_selected',
             'function_selected',
             'aperture_method_selected',
             'previews_temp_disabled')
    def _live_update_marks(self, event={}):
        self._update_marks(event)

    @skip_if_not_tray_instance()
    def _update_marks(self, event={}):
        visible = self.show_live_preview and self.is_active

        if not visible:
            self._clear_marks()
            return

        # ensure the correct visibility, always (whether or not there have been updates)
        if hasattr(self.marks['extract'], 'visible'):
            self.marks['extract'].visible = True

        # _live_update will skip if no updates since last active
        self._live_update_extract(event)

    @skip_if_no_updates_since_last_active()
    @with_temp_disable(timeout=0.4)
    def _live_update_extract(self, event={}):
        self._update_extract()

    @skip_if_not_tray_instance()
    def _update_extract(self):
        try:
            ext = self.extract(add_data=False)
        except (ValueError, Exception):
            self._clear_marks()
            return False

        self.marks['extract'].update_xy(self._preview_x_from_extracted(ext),
                                        self._preview_y_from_extracted(ext))



import os
from pathlib import Path

from astropy.io.registry.base import IORegistryError
from glue_jupyter.common.toolbar_vuetify import read_icon
import ipyvuetify as v
from traitlets import List, Unicode, Dict, Bool, observe

from jdaviz import configs as jdaviz_configs
from jdaviz import __version__
from jdaviz.cli import DEFAULT_VERBOSITY, DEFAULT_HISTORY_VERBOSITY, ALL_JDAVIZ_CONFIGS
from jdaviz.configs.default.plugins.data_tools.file_chooser import FileChooser
from jdaviz.core.data_formats import identify_helper
from jdaviz.core.tools import ICON_DIR
from jdaviz.core.template_mixin import show_widget
from jdaviz.utils import download_uri_to_path

STATUS_HINTS = {
    'idle': "Provide a file path, or pick which viz tool to open",
    'identifying': "Identifying which tool is best to visualize your file...",
    'invalid path': "Error: Cannot find file. Please check your file path.",
    'id ok': "The below tools can best visualize your file. Pick which one you want to use.",
    'id failed': "We couldn’t identify which tool is best for your file. Pick a tool below to use."

}


def open(filename, show=True, **kwargs):
    '''
    Automatically detect the correct configuration based on a given file,
    load the data, and display the configuration

    Parameters
    ----------
    filename : str (path-like)
        Name for a local data file.
    show : bool
        Determines whether to immediately show the application

    All other arguments are interpreted as load_data arguments for
    the autoidentified configuration class

    Returns
    -------
    Jdaviz ConfigHelper : jdaviz.core.helpers.ConfigHelper
        The autoidentified ConfigHelper for the given data
    '''
    # first catch URIs and download them, or return filename unchanged:
    if "local_path" in kwargs:
        fn_dl_kw = {"local_path": kwargs["local_path"]}
    else:
        fn_dl_kw = {}
    filename = download_uri_to_path(filename, cache=True, **fn_dl_kw)

    # Identify the correct config
    compatible_helpers, hdul = identify_helper(filename)
    if len(compatible_helpers) > 1:
        raise NotImplementedError(f"Multiple helpers provided: {compatible_helpers}."
                                  "Unsure which to launch")
    else:
        return _launch_config_with_data(compatible_helpers[0], hdul, filepath=filename,
                                        show=show, **kwargs)


def _launch_config_with_data(config, data=None, show=True, filepath=None, **kwargs):
    '''
    Launch jdaviz with a specific, known configuration and data

    Parameters
    ----------
    config : str (path-like)
        Name for a local data file.
    data : str or any Jdaviz-compatible data
        A filepath or Jdaviz-compatible data object (such as Spectrum1D or CCDData)
    show : bool
        Determines whether to immediately show the application
    filepath : str
        Filepath to use as fallback if ``data`` fails to load.

    All other arguments are interpreted as load_data arguments for
    the autoidentified configuration class

    Returns
    -------
    Jdaviz ConfigHelper : jdaviz.core.helpers.ConfigHelper
        The loaded ConfigHelper with data loaded
    '''
    viz_class = getattr(jdaviz_configs, config.capitalize())

    # Create config instance
    verbosity = kwargs.pop('verbosity', DEFAULT_VERBOSITY)
    history_verbosity = kwargs.pop('history_verbosity', DEFAULT_HISTORY_VERBOSITY)
    viz_helper = viz_class(verbosity=verbosity, history_verbosity=history_verbosity)

    # Load data
    if data not in (None, ''):
        try:
            viz_helper.load_data(data, **kwargs)
        except IORegistryError:
            if filepath is None:
                raise
            viz_helper.load_data(filepath, **kwargs)

    # Display app
    if show:
        viz_helper.show()

    return viz_helper


class Launcher(v.VuetifyTemplate):
    template_file = __file__, "launcher.vue"
    configs = List().tag(sync=True)
    filepath = Unicode().tag(sync=True)
    compatible_configs = List().tag(sync=True)
    config_icons = Dict().tag(sync=True)
    hint = Unicode().tag(sync=True)
    vdocs = Unicode("").tag(sync=True)  # App not available yet, so we need to recompute it here
    # File picker Traitlets
    error_message = Unicode().tag(sync=True)
    valid_path = Bool(True).tag(sync=True)
    file_chooser_visible = Bool(False).tag(sync=True)

    # Define Icons
    cubeviz_icon = Unicode(read_icon(os.path.join(ICON_DIR, 'cubeviz_icon.svg'), 'svg+xml')).tag(sync=True)  # noqa
    specviz_icon = Unicode(read_icon(os.path.join(ICON_DIR, 'specviz_icon.svg'), 'svg+xml')).tag(sync=True)  # noqa
    specviz2d_icon = Unicode(read_icon(os.path.join(ICON_DIR, 'specviz2d_icon.svg'), 'svg+xml')).tag(sync=True)  # noqa
    mosviz_icon = Unicode(read_icon(os.path.join(ICON_DIR, 'mosviz_icon.svg'), 'svg+xml')).tag(sync=True)  # noqa
    imviz_icon = Unicode(read_icon(os.path.join(ICON_DIR, 'imviz_icon.svg'), 'svg+xml')).tag(sync=True)  # noqa

    def __init__(self, main=None, configs=ALL_JDAVIZ_CONFIGS, filepath='',
                 height=None, *args, **kwargs):
        self.vdocs = 'latest' if 'dev' in __version__ else 'v'+__version__

        if main is None:
            main = v.Sheet(class_="mx-25",
                           attributes={"id": "popout-widget-container"},
                           color="#00212C",
                           height=height)

        self.main = main
        self.configs = configs
        self.height = f"{height}px" if isinstance(height, int) else height
        # Set all configs to compatible at first load (for loading blank config)
        self.compatible_configs = configs

        self.config_icons = {
            'cubeviz': self.cubeviz_icon,
            'specviz': self.specviz_icon,
            'specviz2d': self.specviz2d_icon,
            'mosviz': self.mosviz_icon,
            'imviz': self.imviz_icon
        }

        start_path = os.environ.get('JDAVIZ_START_DIR', os.path.curdir)
        self._file_chooser = FileChooser(start_path)
        self.components = {'g-file-import': self._file_chooser}
        self.loaded_data = None
        self.hint = STATUS_HINTS['idle']

        self.filepath = filepath

        super().__init__(*args, **kwargs)

    @observe('filepath')
    def _filepath_changed(self, *args):
        self.hint = STATUS_HINTS['identifying']
        self.compatible_configs = []
        if self.filepath in (None, ''):
            self.compatible_configs = self.configs
            self.loaded_data = None
            self.hint = STATUS_HINTS['idle']
        else:
            path = Path(self.filepath)
            if not path.is_file():
                self.loaded_data = None
                self.hint = STATUS_HINTS['invalid path']
            else:
                try:
                    self.compatible_configs, self.loaded_data = identify_helper(self.filepath)
                    self.hint = STATUS_HINTS['id ok']
                except Exception:
                    self.hint = STATUS_HINTS['id failed']
                    self.compatible_configs = self.configs
                    self.loaded_data = self.filepath
                finally:
                    if len(self.compatible_configs) > 0 and self.loaded_data is None:
                        self.loaded_data = self.filepath
        # Clear hint if it's still stuck on "Identifying". We're in an ambiguous state
        self.hint = ('' if self.hint == STATUS_HINTS['identifying'] else self.hint)

    def vue_choose_file(self, *args, **kwargs):
        if self._file_chooser.file_path is None:
            self.error_message = "No file selected"
        elif Path(self._file_chooser.file_path).is_file():
            self.file_chooser_visible = False
            self.filepath = self._file_chooser.file_path

    def vue_launch_config(self, event):
        config = event.get('config')
        helper = _launch_config_with_data(config, self.loaded_data,
                                          filepath=self.filepath, show=False)
        if self.height not in ['100%', '100vh']:
            # We're in jupyter mode. Set to default height
            default_height = helper.app.state.settings['context']['notebook']['max_height']
            helper.app.layout.height = default_height
            self.main.height = default_height
        self.main.color = 'transparent'
        self.main.children = [helper.app]

    @property
    def main_with_launcher(self):
        self.main.children = [self]
        return self.main


def show_launcher(configs=ALL_JDAVIZ_CONFIGS, filepath='', height='450px'):
    '''Display an interactive Jdaviz launcher to select your data and compatible configuration

    Parameters
    ----------
    filepath: str, optional
        The path to the file to load into Jdaviz. Will autopopulate the File Path field

    height: int, optional
        The height of the top-level application widget, in pixels. Will be passed and processed
        by the selected configuration. Applies to the launcher and all instances of the same
        application in the notebook.
    '''
    # Color defined manually due to the custom theme not being defined yet (in main_styles.vue)
    height = f"{height}px" if isinstance(height, int) else height
    launcher = Launcher(None, configs, filepath, height)
    show_widget(launcher.main_with_launcher, loc='inline', title=None)



from PyInstaller.utils.hooks import (collect_submodules,
                                     collect_data_files,
                                     copy_metadata)

datas = collect_data_files('jupyter_client')
datas += copy_metadata('jupyter_client')
hiddenimports = collect_submodules("jupyter_client")



import numpy as np
import pytest
from astropy import units as u
from astropy.nddata import NDData
import matplotlib
from numpy.testing import assert_allclose
from photutils.datasets import make_4gaussians_image

from jdaviz.configs.default.plugins.plot_options.plot_options import SplineStretch


@pytest.mark.filterwarnings('ignore')
def test_multiselect(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube)
    po = cubeviz_helper.app.get_tray_item_from_name('g-plot-options')

    # default selection for viewer should be flux-viewer (first in list) and nothing for layer
    assert po.viewer.multiselect is False
    assert po.layer.multiselect is False
    assert po.viewer.selected == 'flux-viewer'
    assert po.layer.selected == 'Unknown spectrum object[FLUX]'

    # deprecated functionality to toggle both at once, replace with setting
    # individually when removing
    po.multiselect = True
    assert po.viewer.multiselect is True
    assert po.layer.multiselect is True

    po.viewer.selected = ['flux-viewer', 'uncert-viewer']
    # any internal call to reset the defaults (due to a change in the choices by adding or removing
    # a viewer, etc) should not reset the selection
    po.viewer._apply_default_selection(skip_if_current_valid=True)
    assert po.viewer.selected == ['flux-viewer', 'uncert-viewer']

    # from API could call po.axes_visible.value = False, but we'll also test the vue-level wrapper
    po.vue_set_value({'name': 'axes_visible_value', 'value': False})
    assert po.axes_visible.value is False

    for viewer_name in ['flux-viewer', 'uncert-viewer']:
        assert cubeviz_helper.app.get_viewer(viewer_name).state.show_axes is False

    assert cubeviz_helper.app.get_viewer('spectrum-viewer').state.show_axes is True
    assert po.axes_visible.sync['mixed'] is False

    # adding another viewer should show a mixed-state
    po.viewer.selected = ['flux-viewer', 'uncert-viewer', 'spectrum-viewer']
    # spectrum viewer is excluded from the axes_visible/show_axes logic
    assert len(po.axes_visible.linked_states) == 2
    assert len(po.axes_visible.sync['icons']) == 2
    assert po.axes_visible.sync['mixed'] is False

    # from API could call po.axes_visible.unmix_state, but we'll also test the vue-level wrapper
    po.vue_unmix_state('axes_visible')
    assert po.axes_visible.sync['mixed'] is False
    assert po.axes_visible.value is False

    for viewer_name in ['flux-viewer', 'uncert-viewer']:
        assert cubeviz_helper.app.get_viewer(viewer_name).state.show_axes is False

    for viewer_name in ['spectrum-viewer']:
        assert cubeviz_helper.app.get_viewer(viewer_name).state.show_axes is True

    po.viewer_multiselect = False
    # should default back to the first selected entry
    assert po.viewer.multiselect is False
    assert po.viewer.selected == 'flux-viewer'
    po.layer_multiselect = False
    assert po.axes_visible.value is False

    po.viewer.selected = 'spectrum-viewer'
    assert po.axes_visible.sync['in_subscribed_states'] is False
    assert len(po.axes_visible.linked_states) == 0


@pytest.mark.filterwarnings('ignore')
def test_stretch_histogram(cubeviz_helper, spectrum1d_cube_with_uncerts):
    cubeviz_helper.load_data(spectrum1d_cube_with_uncerts)
    po = cubeviz_helper.app.get_tray_item_from_name('g-plot-options')
    po.plugin_opened = True

    assert po.stretch_histogram is not None

    # check the colorbar
    cb = po.stretch_histogram._marks["colorbar"]
    assert_allclose(cb.x, po.stretch_histogram.figure.marks[0].x)
    assert_allclose(cb.y, 1)
    assert cb.colors == [  # Gray scale, linear
        '#050505', '#0f0f0f', '#191919', '#232323', '#2e2e2e',
        '#383838', '#424242', '#4c4c4c', '#575757', '#616161',
        '#6b6b6b', '#757575', '#808080', '#8a8a8a', '#949494',
        '#9e9e9e', '#a8a8a8', '#b3b3b3', '#bdbdbd', '#c7c7c7',
        '#d1d1d1', '#dcdcdc', '#e6e6e6', '#f0f0f0', '#fafafa']

    hist_lyr = po.stretch_histogram.layers['histogram']
    flux_cube_sample = hist_lyr.layer.data['x']

    # changing viewer should change results
    po.viewer.selected = 'uncert-viewer'
    with pytest.raises(AssertionError):
        assert_allclose(hist_lyr.layer.data['x'], flux_cube_sample)

    po.viewer.selected = 'flux-viewer'
    assert_allclose(hist_lyr.layer.data['x'], flux_cube_sample)

    # change viewer limits
    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    fv.state.x_max = 0.5 * fv.state.x_max
    # viewer limits should not be affected by default
    # (re-retrieve layer - it should not have changed)
    hist_lyr = po.stretch_histogram.layers['histogram']
    assert_allclose(hist_lyr.layer.data['x'], flux_cube_sample)

    # set to listen to viewer limits, the length of the samples will change
    # (in this case the layer itself has been replaced)
    po.stretch_hist_zoom_limits = True
    hist_lyr = po.stretch_histogram.layers['histogram']
    assert len(hist_lyr.layer.data['x']) != len(flux_cube_sample)

    po.stretch_vmin.value = 0.5
    po.stretch_vmax.value = 1

    assert po.stretch_histogram.marks['vmin'].x[0] == po.stretch_vmin.value
    assert po.stretch_histogram.marks['vmax'].x[0] == po.stretch_vmax.value

    assert po.stretch_histogram.viewer.state.hist_n_bin == 25
    po.stretch_hist_nbins = 20
    assert po.stretch_histogram.viewer.state.hist_n_bin == 20

    po.set_histogram_limits(x_min=0.25, x_max=2)
    assert po.stretch_histogram.viewer.state.x_min == 0.25
    assert po.stretch_histogram.viewer.state.x_max == 2

    po.set_histogram_limits(y_min=1, y_max=2)
    assert po.stretch_histogram.viewer.state.y_min == 1
    assert po.stretch_histogram.viewer.state.y_max == 2

    assert po.stretch_histogram.marks['vmin'].x[0] == po.stretch_vmin.value
    assert po.stretch_histogram.marks['vmax'].x[0] == po.stretch_vmax.value

    # Make sure some tools work

    po_panzoom = po.stretch_histogram.toolbar.tools["jdaviz:panzoom"]
    po_panzoom.activate()
    po_panzoom.deactivate()

    po_prevzoom = po.stretch_histogram.toolbar.tools["jdaviz:prevzoom"]
    po_prevzoom.activate()


@pytest.mark.filterwarnings('ignore')
def test_user_api(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube)
    po = cubeviz_helper.plugins['Plot Options']

    assert po.multiselect is False
    assert "multiselect" in po.viewer.__repr__()

    # regression test for https://github.com/spacetelescope/jdaviz/pull/1708
    # user calls to select_default should revert even if current entry is valid
    po.viewer = 'spectrum-viewer'
    assert po.viewer.selected == 'spectrum-viewer'
    po.viewer.select_default()
    assert po.viewer.selected == 'flux-viewer'

    po.select_all()
    assert po.multiselect is True
    assert len(po.viewer.selected) == 3

    po.viewer.select_none()
    assert len(po.viewer.selected) == 0

    po.viewer.select_default()
    po.layer.select_default()
    assert po.multiselect is True
    assert len(po.viewer.selected) == 1
    assert len(po.layer.selected) == 1

    with pytest.raises(ValueError):
        po.viewer = ['flux-viewer', 'invalid-viewer']

    assert len(po.viewer.selected) == 1

    po.viewer = 'flux-viewer'
    po.layer.select_all()

    # check a plot option with and without choices
    assert hasattr(po.stretch_preset, 'choices')
    assert len(po.stretch_preset.choices) > 1
    assert "choices" in po.stretch_preset._obj.__repr__()
    assert not hasattr(po.image_contrast, 'choices')
    assert "choices" not in po.image_contrast._obj.__repr__()

    # try setting with both label and value
    po.stretch_preset = 90
    po.stretch_preset = 'Min/Max'

    # try eq on both text and value
    assert po.image_colormap == 'gray'
    assert po.image_colormap == 'Gray'

    # toggle contour (which has a spinner implemented)
    po.contour_visible = True
    assert po._obj.contour_spinner is False


def test_stretch_spline(imviz_helper):
    image_1 = NDData(make_4gaussians_image(), unit=u.nJy)
    # Load the test data into imviz
    imviz_helper.load_data(image_1)
    po = imviz_helper.plugins['Plot Options']

    # Configure initial stretch options and select "Spline" function
    with po.as_active():
        po.stretch_vmin.value = 10
        po.stretch_vmax.value = 100
        po.stretch_curve_visible = True
        po.stretch_function = "Spline"

    # Retrieve knots data from the generated histogram
    scatter_obj = po._obj.stretch_histogram.marks["stretch_knots"]
    knots_x = scatter_obj.x
    knots_y = scatter_obj.y

    # Expected knots based on initial stretch settings
    expected_x = [10., 19., 28., 73., 100.]
    expected_y = np.array([0., 0.05, 0.3, 0.9, 1.]) * 0.9  # rescaled to 0.9

    # Validate if the generated knots match the expected knots
    assert scatter_obj.visible
    assert_allclose(knots_x, expected_x)
    assert_allclose(knots_y, expected_y)

    # Update the stretch options to new values and verify the knots update correctly
    with po.as_active():
        po.stretch_vmin.value = 10
        po.stretch_vmax.value = 80
        po.stretch_function = "Spline"

    knots_x = scatter_obj.x
    knots_y = scatter_obj.y

    # Expected knots based on updated stretch settings
    expected_x = [10., 17., 24., 59., 80.]
    expected_y = np.array([0., 0.05, 0.3, 0.9, 1.]) * 0.9  # rescaled to 0.9

    # Validate if the generated knots for updated settings match the expected values
    assert scatter_obj.visible
    assert_allclose(knots_x, expected_x)
    assert_allclose(knots_y, expected_y)

    # Disable the stretch curve and ensure no knots or stretches are visible
    with po.as_active():
        po.stretch_curve_visible = False
    stretch_curve = po._obj.stretch_histogram.marks['stretch_curve']
    assert len(stretch_curve.y) == 0
    assert len(stretch_curve.x) == 0
    assert len(scatter_obj.x) == 0
    assert len(scatter_obj.x) == 0


def test_update_knots_mismatched_length():
    stretch = SplineStretch()
    with pytest.raises(ValueError):
        stretch.update_knots(x=[0, 0.1, 0.2], y=[0, 0.05])


def test_apply_presets(imviz_helper):
    arr = np.arange(36).reshape(6, 6)
    po = imviz_helper.plugins['Plot Options']

    preset_colors_4 = ["#0000ff", "#00ff00", po._obj.swatches_palette[1][0].lower(),
                       po._obj.swatches_palette[0][0].lower()]

    # Test applying presets with < 6 layers

    for i in range(4):
        imviz_helper.load_data(arr, data_label=f"array_{i}")

    po.image_color_mode = "Color"
    po.apply_RGB_presets()

    for i in range(4):
        po.layer = f"array_{i}"
        assert po.stretch_function.value == "arcsinh"
        assert po.stretch_preset.value == 99
        assert po.image_color.value == preset_colors_4[i]

    # Test applying presets with > 5 layers

    for i in range(4):
        imviz_helper.load_data(arr, data_label=f"array_{i+4}")

    po.layer = "array_5"
    po.image_visible = False
    po.layer = "array_3"

    po.apply_RGB_presets()

    assert po.layer.selected == "array_3"
    assert po.stretch_function.value == "arcsinh"
    po.layer = "array_5"
    # Make sure this one didn't change
    assert po.stretch_function.value == "linear"

    # Turn layer 5 back on
    po.image_visible = True
    po.apply_RGB_presets()

    colorbar_colors = matplotlib.colormaps['gist_rainbow'].resampled(8)
    color_ind = 0

    def _rgb_to_hex(rgb):
        rgb = [int(x * 255) for x in rgb]
        return f"#{rgb[0]:02x}{rgb[1]:02x}{rgb[2]:02x}{rgb[3]:02x}"

    for i in range(8):
        po.layer = f"array_{i}"
        assert po.stretch_function.value == "arcsinh"
        assert po.stretch_preset.value == 99
        assert po.image_color.value == matplotlib.colors.to_hex(colorbar_colors(7-color_ind),
                                                                keep_alpha=True)
        color_ind += 1


def test_track_mixed_states(imviz_helper):
    # Initialize two viewer with 3 data each.
    # Each layer of the data will be RGB
    arr = np.arange(36).reshape(6, 6)
    po = imviz_helper.app.get_tray_item_from_name("g-plot-options")
    rgb_colors = ["#ff0000", "#00ff00", "#0000ff"]

    for i in range(3):
        imviz_helper.load_data(arr, data_label=f"array_{i}")

    po.image_color_mode_value = 'One color per layer'
    for i in range(3):
        po.layer_selected = f"array_{i}"
        po.image_color.value = rgb_colors[i]

    imviz_helper.create_image_viewer(viewer_name="imviz-1")
    po.viewer_selected = "imviz-1"
    po.image_color_mode_value = 'One color per layer'
    for i in range(3):
        imviz_helper.app.add_data_to_viewer("imviz-1", data_label=f"array_{i}")
        po.layer_selected = f"array_{i}"
        po.image_color.value = rgb_colors[i]

    # Switch to multiselect to test mixing and unmixing of states
    po.viewer_multiselect = True
    po.layer_multiselect = True
    po.viewer_selected = ["imviz-0", "imviz-1"]
    assert po.layer.items[-1]["label"] == "array_2"
    # The corresponding layer in each viewer is the same color,
    # so the state is not mixed.
    assert len(po.layer.items[-1]["colors"]) == 1

    # Change the color of one of the layers in one viewer
    po.viewer_selected = ["imviz-1"]
    po.layer_selected = ["array_2"]
    po.image_color.value = "#595959"
    po.viewer_selected = ["imviz-0", "imviz-1"]
    # The color state is now mixed when two viewers are selected
    assert po.layer.items[-1]["label"] == "array_2"
    assert len(po.layer.items[-1]["colors"]) == 2

    # Now test mixed visibility
    po.viewer_selected = ["imviz-1"]
    po.layer_selected = ["array_1", "array_2"]
    po.image_visible.value = False
    po.viewer_selected = ["imviz-0", "imviz-1"]
    assert po.layer.items[-1]["label"] == "array_2"
    assert po.layer.items[-1]["visible"] == 'mixed'
    assert po.layer.items[-2]["label"] == "array_1"
    assert po.layer.items[-2]["visible"] == 'mixed'

    # Test unmixing visibility
    po.image_visible.unmix_state(True)
    assert po.layer.items[-1]["visible"] is True
    assert po.layer.items[-1]["visible"]
    assert po.layer.items[-2]["visible"] is True
    assert po.layer.items[-2]["visible"]

    # Now test unmixing color
    po.viewer_selected = ["imviz-0", "imviz-1"]
    assert len(po.layer.items[-1]["colors"]) > 1
    assert len(po.layer.items[-2]["colors"]) == 1

    # Make sure that all selected layers are no longer
    # mixed state and are the same color
    po.image_color.unmix_state()
    assert len(po.layer.items[-1]["colors"]) == 1
    assert len(po.layer.items[-2]["colors"]) == 1
    assert po.image_color.value == "#00ff00"
    assert po.layer.items[-1]["colors"][0] == "#00ff00"
    assert po.layer.items[-2]["colors"][0] == "#00ff00"

    # test spline stretch mixed state
    po.viewer_selected = ["imviz-0"]
    po.layer_selected = ["array_1"]
    po.stretch_function_value = 'spline'
    po.viewer_selected = ["imviz-0", "imviz-1"]
    assert po.stretch_function_sync['mixed']

    po.vue_unmix_state(['stretch_function', 'stretch_params'])
    assert not po.stretch_function_sync['mixed']
    assert po.stretch_function_value == 'spline'
    assert not po.stretch_params_sync['mixed']


def test_segmentation_image(imviz_helper):
    # Make one circular segment for a hypothetical
    # image with one source:
    nx = ny = 100
    radius = 15
    x0 = y0 = 50

    segmentation_map = np.zeros((nx, ny))
    xx, yy = np.meshgrid(np.arange(nx), np.arange(ny))
    in_circle = np.hypot(xx - x0, yy - y0) < radius
    segmentation_map[in_circle] = 1

    imviz_helper.load_data(segmentation_map)

    plot_opts = imviz_helper.plugins['Plot Options']
    plot_opts.image_colormap = 'Random'

    # ensure that stretch preset is adjusted when the image segmentaiton
    # option is selected for the "Random" colormap, so that all colors
    # are uniquely displayed:
    plot_opts._obj.image_segmentation_map_presets()
    assert plot_opts.stretch_preset.value == 100
    assert plot_opts.stretch_function.value == 'linear'
    assert plot_opts.image_bias.value == 0.5
    assert plot_opts.image_contrast.value == 1.0


def test_imviz_select_all_layers(imviz_helper):
    """
    Test to catch a (fixed) bug that was revealed when trying to select
    all layers when data is float32. This was caused when trying to set
    `stretch_vmin_value`.
    """

    arr = np.arange(36.).reshape(6, 6).astype(np.float32)

    # load three images in one viewer
    with imviz_helper.batch_load():
        for i in range(3):
            imviz_helper.load_data(arr, data_label=f"data_{i}")

    plot_options = imviz_helper.plugins['Plot Options']

    plot_options.layer.multiselect = True
    plot_options.select_all()

    # all layers selected, set stretch function to log for all
    plot_options.stretch_function = 'log'

    # and make sure each layer picked up this change
    for layer in plot_options.image_colormap.linked_states:
        assert layer.as_dict()['stretch'] == 'log'



import warnings

import pytest
from astropy.table import Table
from astropy.wcs import WCS
from glue.core.link_helpers import LinkSame
from glue.plugins.wcs_autolinking.wcs_autolinking import AffineLink, OffsetLink, WCSLink
from numpy.testing import assert_allclose
from regions import PixCoord, CirclePixelRegion, PolygonPixelRegion

from jdaviz.configs.imviz.helper import get_reference_image_data
from jdaviz.configs.imviz.tests.utils import (
    BaseImviz_WCS_NoWCS, BaseImviz_WCS_WCS, BaseImviz_WCS_GWCS, BaseImviz_GWCS_GWCS)


class BaseLinkHandler:

    def check_all_pixel_links(self):
        links = self.imviz.app.data_collection.external_links
        assert len(links) == 2
        assert all([isinstance(link, LinkSame) for link in links])

    def check_all_wcs_links(self):
        links = self.imviz.app.data_collection.external_links
        assert len(links) == 3
        assert all([isinstance(link, (AffineLink, OffsetLink)) for link in links])

    def test_pixel_linking(self):
        self.imviz.link_data(align_by='pixels')
        self.check_all_pixel_links()

    @property
    def default_viewer_limits(self):
        return (self.imviz.default_viewer._obj.state.x_min,
                self.imviz.default_viewer._obj.state.x_max,
                self.imviz.default_viewer._obj.state.y_min,
                self.imviz.default_viewer._obj.state.y_max)


class TestLink_WCS_NoWCS(BaseImviz_WCS_NoWCS, BaseLinkHandler):

    def test_wcslink_fallback_pixels(self):
        self.imviz.link_data(align_by='wcs')

        assert self.viewer.get_alignment_method('has_wcs[SCI,1]') == 'wcs'

        # Also check the coordinates display: Last loaded is on top.

        label_mouseover = self.imviz.app.session.application._tools['g-coords-info']
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove', 'domain': {'x': 0, 'y': 0}})

        assert label_mouseover.as_text() == (
            'Pixel x=00.3 y=00.2 Value +0.00000e+00',
            'World 22h30m04.8496s -20d49m59.7490s (ICRS)',
            '337.5202064976 -20.8332636155 (deg)'
        )


class TestLink_WCS_FakeWCS(BaseImviz_WCS_NoWCS, BaseLinkHandler):

    def test_badwcs_no_crash(self):
        # There is WCS but it is non-celestial
        self.imviz.app.data_collection[1].coords = WCS(naxis=2)

        self.check_all_pixel_links()

        assert self.viewer.get_alignment_method('has_wcs[SCI,1]') == 'self'
        assert self.viewer.get_alignment_method('no_wcs[SCI,1]') == 'pixels'

        # Also check the coordinates display: Last loaded is on top.

        label_mouseover = self.imviz.app.session.application._tools['g-coords-info']
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 0, 'y': 0}})
        assert label_mouseover.as_text() == ('Pixel x=00.0 y=00.0 Value +0.00000e+00', '', '')

        # blink image through keypress
        self.viewer.on_mouse_or_key_event({'event': 'keydown', 'key': 'b',
                                           'domain': {'x': 0, 'y': 0}})
        assert label_mouseover.as_text() == ('Pixel x=00.0 y=00.0 Value +0.00000e+00',
                                             'World 22h30m04.8674s -20d49m59.9990s (ICRS)',
                                             '337.5202808000 -20.8333330600 (deg)')


class TestLink_WCS_WCS(BaseImviz_WCS_WCS, BaseLinkHandler):

    def test_wcslink_affine_with_extras(self):
        orig_pixel_limits = self.default_viewer_limits
        assert_allclose(orig_pixel_limits, (-0.5, 9.5, -0.5, 9.5))

        self.imviz.link_data(align_by='wcs', wcs_fallback_scheme=None)
        links = self.imviz.app.data_collection.external_links
        assert len(links) == 2
        assert isinstance(links[0], (AffineLink, OffsetLink))
        assert self.viewer.get_alignment_method('has_wcs_2[SCI,1]') == 'wcs'

        # Customize display on second image (last loaded).
        self.viewer.set_colormap('Viridis')
        self.viewer.stretch = 'sqrt'
        self.viewer.cuts = (0, 100)

        # Add subsets
        self.imviz.plugins['Subset Tools'].import_region([
            CirclePixelRegion(center=PixCoord(x=2.55, y=3.55), radius=1.05),
            CirclePixelRegion(center=PixCoord(x=6, y=2), radius=5).to_sky(self.wcs_1),
            PolygonPixelRegion(vertices=PixCoord(x=[1, 2, 2], y=[1, 1, 2])).to_sky(self.wcs_1),
            PolygonPixelRegion(vertices=PixCoord(x=[2, 3, 3], y=[2, 2, 3])).to_sky(self.wcs_1)
        ], combination_mode="new")

        # Add markers.
        tbl = Table({'x': (0, 0), 'y': (0, 1)})
        self.viewer.add_markers(tbl, marker_name='xy_markers')
        assert 'xy_markers' in self.imviz.app.data_collection.labels

        # Ensure display is still customized.
        assert self.viewer.state.layers[1].cmap.name == 'viridis'
        assert self.viewer.state.layers[1].stretch == 'sqrt'
        assert_allclose((self.viewer.state.layers[1].v_min, self.viewer.state.layers[1].v_max),
                        (0, 100))

        # Ensure subsets are still there.
        all_labels = [layer.layer.label for layer in self.viewer.state.layers]
        # Retrieved subsets as sky regions from Subset plugin, and ensure they
        # match what was loaded and that they are in sky coordinates.
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore',
                                    message='Regions skipped: MaskedSubset 1, MaskedSubset 2')
            subset_as_regions = self.imviz.plugins['Subset Tools'].get_regions()
        assert sorted(subset_as_regions) == ['Subset 1', 'Subset 2']
        assert_allclose(subset_as_regions['Subset 1'].center.ra.deg, 337.519449)
        assert_allclose(subset_as_regions['Subset 2'].center.ra.deg, 337.518498)
        # ensure agreement between app.get_subsets and subset_tools.get_regions
        ss = self.imviz.app.get_subsets(include_sky_region=True)
        assert ss['Subset 1'][0]['sky_region'] == subset_as_regions['Subset 1']
        assert ss['Subset 2'][0]['sky_region'] == subset_as_regions['Subset 2']

        assert 'MaskedSubset 1' in all_labels
        assert 'MaskedSubset 2' in all_labels

        # Markers should still exist since the type has not changed
        # Zoom and pan will reset in this case, so we do not check those.
        assert 'xy_markers' in self.imviz.app.data_collection.labels
        assert len(self.viewer._marktags) == 1

        # Pan/zoom.
        self.viewer.center_on((5, 5))
        self.viewer.zoom_level = 0.789
        ans = (self.viewer.state.x_min, self.viewer.state.y_min,
               self.viewer.state.x_max, self.viewer.state.y_max)

        # Ensure pan/zoom does not change when markers are not present.
        assert_allclose((self.viewer.state.x_min, self.viewer.state.y_min,
                        self.viewer.state.x_max, self.viewer.state.y_max), ans)

        # Also check the coordinates display: Last loaded is on top.

        label_mouseover = self.imviz.app.session.application._tools['g-coords-info']
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 0, 'y': 0}})

        lmtext = label_mouseover.as_text()
        assert lmtext[0] == 'Pixel x=01.3 y=00.2 Value +1.00000e+00'
        assert lmtext[1:] == ('World 22h30m04.8496s -20d49m59.7490s (ICRS)',
                              '337.5202064976 -20.8332636155 (deg)')

        # blink image through clicking with blink tool
        self.viewer.toolbar.active_tool_id = 'jdaviz:blinkonce'
        self.viewer.toolbar.active_tool.on_click({'event': 'click', 'domain': {'x': 0, 'y': 0}})
        assert label_mouseover.as_text()[0] == 'Pixel x=00.3 y=00.2 Value +1.00000e+00'
        assert label_mouseover.as_text()[1:] == ('World 22h30m04.8496s -20d49m59.7490s (ICRS)',
                                                 '337.5202064976 -20.8332636155 (deg)')

        # Changing link type will raise an error
        with pytest.raises(ValueError, match=".*only be changed after existing subsets are deleted"):  # noqa: E501
            self.imviz.link_data(align_by='pixels', wcs_fallback_scheme=None)

        self.viewer.reset_markers()
        self.imviz.plugins["Orientation"].delete_subsets()
        self.imviz.link_data(align_by='pixels', wcs_fallback_scheme=None)
        assert 'xy_markers' not in self.imviz.app.data_collection.labels
        assert len(self.viewer._marktags) == 0

    def test_wcslink_fullblown(self):
        self.imviz.link_data(align_by='wcs', wcs_fallback_scheme=None,
                             wcs_fast_approximation=False)
        links = self.imviz.app.data_collection.external_links
        assert len(links) == 2
        assert isinstance(links[0], WCSLink)
        assert self.viewer.get_alignment_method('has_wcs_1[SCI,1]') == 'wcs'
        assert self.viewer.get_alignment_method('has_wcs_2[SCI,1]') == 'wcs'

    # Also test other exception handling here.

    def test_invalid_inputs(self):
        with pytest.raises(KeyError):
            self.imviz.link_data(align_by='foo')

        with pytest.raises(ValueError, match='not found in data collection external links'):
            self.viewer.get_alignment_method('foo')


class TestLink_WCS_GWCS(BaseImviz_WCS_GWCS):

    def test_wcslink_rotated(self):
        # FITS WCS will be reference, GWCS is rotated, no_wcs linked by pixel to ref.
        self.imviz.link_data(align_by='wcs')

        # The zoom box for GWCS is now a rotated rombus.
        fits_wcs_zoom_limits = self.viewer._get_zoom_limits(
            self.imviz.app.data_collection['fits_wcs[DATA]'])
        gwcs_zoom_limits = self.viewer._get_zoom_limits(
            self.imviz.app.data_collection['gwcs[DATA]'])

        # x_min, y_min
        # x_min, y_max
        # x_max, y_max
        # x_max, y_min
        assert_allclose(fits_wcs_zoom_limits,
                        [[-2.406711, -1.588057],
                         [-2.697746, 11.137127],
                         [10.148055, 10.554429],
                         [10.439091, -2.170755]], rtol=1e-5)
        assert_allclose(gwcs_zoom_limits,
                        [[2.636299, 12.732915],
                         [13.375281, 5.007547],
                         [6.300587, -5.126264],
                         [-4.438394, 2.599103]], rtol=1e-5)

        # Also check the coordinates display: Last loaded is on top.
        # Cycle order: GWCS, FITS WCS
        label_mouseover = self.imviz.app.session.application._tools['g-coords-info']
        xy = self.viewer._get_real_xy(self.imviz.app.data_collection[0], 0, 0, reverse=True)
        label_mouseover._viewer_mouse_event(
            self.viewer, {'event': 'mousemove', 'domain': {'x': xy[0], 'y': xy[1]}})
        assert label_mouseover.as_text() == ('Pixel x=02.7 y=09.8',
                                             'World 00h14m19.6141s -30d23m31.4091s (ICRS)',
                                             '3.5817255823 -30.3920580740 (deg)')
        assert not label_mouseover.row1_unreliable
        assert not label_mouseover.row2_unreliable
        assert not label_mouseover.row3_unreliable

        # Make sure GWCS now can extrapolate.
        xy = self.viewer._get_real_xy(self.imviz.app.data_collection[1], -1, -1, reverse=True)
        label_mouseover._viewer_mouse_event(
            self.viewer, {'event': 'mousemove', 'domain': {'x': xy[0], 'y': xy[1]}})
        assert label_mouseover.as_text() == ('Pixel x=-1.0 y=-1.0',
                                             'World 00h14m19.5829s -30d23m30.9860s (ICRS)',
                                             '3.5815955408 -30.3919405616 (deg)')
        # FITS WCS is reference data and has no concept of bounding box
        # but cursor is outside GWCS bounding box
        assert label_mouseover.row1_unreliable
        assert label_mouseover.row2_unreliable
        assert label_mouseover.row3_unreliable

        xy = self.viewer._get_real_xy(self.imviz.app.data_collection[0], 0, 0, reverse=True)
        self.viewer.blink_once()
        label_mouseover._viewer_mouse_event(
            self.viewer, {'event': 'mousemove', 'domain': {'x': xy[0], 'y': xy[1]}})
        assert label_mouseover.as_text()[0] in (
            'Pixel x=00.0 y=00.0 Value +1.00000e+00 electron / s',
            'Pixel x=-0.0 y=00.0 Value +1.00000e+00 electron / s',
            'Pixel x=00.0 y=-0.0 Value +1.00000e+00 electron / s',
            'Pixel x=-0.0 y=-0.0 Value +1.00000e+00 electron / s'
        )
        assert label_mouseover.as_text()[1:] == ('World 00h14m19.6141s -30d23m31.4091s (ICRS)',
                                                 '3.5817255823 -30.3920580740 (deg)')
        assert not label_mouseover.row1_unreliable
        assert not label_mouseover.row2_unreliable
        assert not label_mouseover.row3_unreliable

        # Regression test for https://github.com/spacetelescope/jdaviz/issues/2079 to
        # make sure this does not crash.
        viewer2 = self.imviz.create_image_viewer(viewer_name='second')
        viewer2.state.reset_limits()


class TestLink_GWCS_GWCS(BaseImviz_GWCS_GWCS):

    def test_pixel_linking(self):
        self.imviz.link_data(align_by='pixels')

        # Check the coordinates display: Last loaded is on top.
        label_mouseover = self.imviz.app.session.application._tools['g-coords-info']
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove', 'domain': {'x': -1, 'y': -1}})
        assert label_mouseover.as_text() == ('Pixel x=-1.0 y=-1.0',
                                             'World 00h14m19.5987s -30d23m31.0683s (ICRS)',
                                             '3.5816611274 -30.3919634282 (deg)')
        assert not label_mouseover.row1_unreliable
        assert label_mouseover.row2_unreliable
        assert label_mouseover.row3_unreliable

        # Back to reference image with bounds check.
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'keydown', 'key': 'b',
                                             'domain': {'x': -1, 'y': -1}})
        self.viewer.on_mouse_or_key_event({'event': 'keydown', 'key': 'b',
                                           'domain': {'x': -1, 'y': -1}})
        assert label_mouseover.as_text() == ('Pixel x=-1.0 y=-1.0',
                                             'World 00h14m19.5829s -30d23m30.9860s (ICRS)',
                                             '3.5815955408 -30.3919405616 (deg)')
        assert not label_mouseover.row1_unreliable
        assert label_mouseover.row2_unreliable
        assert label_mouseover.row3_unreliable


def test_imviz_no_data(imviz_helper):
    refdata, iref = get_reference_image_data(imviz_helper.app)
    assert refdata is None
    assert iref == -1

    imviz_helper.link_data()  # Just no-op, do not crash
    links = imviz_helper.app.data_collection.external_links
    assert len(links) == 0

    with pytest.raises(ValueError, match='No reference data for link look-up'):
        imviz_helper.default_viewer._obj.get_alignment_method('foo')



from .viewers import *  # noqa
from .gaussian_smooth.gaussian_smooth import *  # noqa
from .data_tools import *  # noqa
from .viewer_creator import *  # noqa
from .subset_tools.subset_tools import *  # noqa
from .subset_widget.subset_widget import * # noqa
from .model_fitting.model_fitting import *  # noqa
from .collapse.collapse import *  # noqa
from .line_lists.line_lists import *  # noqa
from .metadata_viewer.metadata_viewer import *  # noqa
from .export.export import *  # noqa
from .plot_options.plot_options import *  # noqa
from .markers.markers import * # noqa
from .data_quality.data_quality import *  # noqa
from .about.about import *  # noqa
from .virtual_observatory import *  # noqa
